'''
This .py-file is used to do the machine learning for the publication: 

"Advancing Data Quality Assurance with Machine Learning: A Case Study on Wind Vane Stalling Detection"

Authors: Vincent S. de Feiter, Jessica M.I. Strickland and Irene Garcia Marti | Contact: vincent.defeiter@wur.nl 

'''


#%%
#==============================================================================

#                        I M P O R T   M O D U L E S

#==============================================================================
import pandas as pd                                                 #work with the data
import numpy as np                                                  #work and modify data arrays
import matplotlib.pyplot as plt                                     #visualisation
from matplotlib.ticker import AutoMinorLocator
import seaborn as sns
from mpl_toolkits.mplot3d import Axes3D
import datetime
import json

from sklearn.model_selection import train_test_split                #splitting the data in train and test data
from sklearn.preprocessing import MinMaxScaler, StandardScaler      #Feature Scaling
from sklearn.metrics import classification_report                   #Metrics: classification report. Build a text report showing the main classification metrics.
from sklearn.metrics import accuracy_score 
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import make_scorer
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import f1_score
from sklearn.metrics import confusion_matrix
from collections import Counter

from sklearn.feature_selection import mutual_info_classif           #Mutual information --> Feature Extraction. 
from sklearn.model_selection import GridSearchCV                    #Gridsearch (for Model Tuning)

#Selected models
from sklearn.svm import OneClassSVM                                 #OneClass Support Vector Machine: binary approach, only with the majority class. 
from sklearn.svm import SVC                                         #Support Vector Machine: binary approach, with the majority and minority class. 
from sklearn.linear_model import LogisticRegression                 #Logistic Regression
from sklearn.neighbors import KNeighborsClassifier                  #K-Neighbors
from sklearn.naive_bayes import GaussianNB                          #Naïve Bayes
from sklearn.ensemble import RandomForestClassifier                 #Random Forest Classifier (multiple decison trees)

#General settings for visualisation
#Use LATEX font
plt.rcParams['text.usetex'] = True
plt.rcParams['text.latex.preamble'] = r'\usepackage{bm}'
plt.rcParams.update({
    "font.family": "serif",
    "font.serif": ["Arial"],
})


fontsize = 15
plt.rcParams['font.size'] = fontsize

# # Update the font sizes for ylabel, xlabel, and tick labels
plt.rcParams['axes.labelsize'] = fontsize   # X and Y labels

# # Optionally, you can scale other elements, such as legend size, here
plt.rcParams['legend.fontsize'] = fontsize
plt.rcParams['axes.titlepad'] = 20   # Title padding
plt.rcParams['axes.labelpad'] = 20   # Axis label padding
plt.rcParams['xtick.major.pad']= 5 
plt.rcParams['ytick.major.pad']= 5

plt.rcParams['axes.edgecolor'] = 'k'  # Color of the plot border
plt.rcParams['axes.linewidth'] = 1  # Width of the plot border (spines)
plt.rcParams['xtick.major.width'] = 1  # Width of major ticks on the x-axis
plt.rcParams['ytick.major.width'] = 1  # Width of major ticks on the y-axis
plt.rcParams['xtick.minor.width'] = 1  # Width of minor ticks on the x-axis
plt.rcParams['ytick.minor.width'] = 1  # Width of minor ticks on the y-axis
plt.rcParams['xtick.major.size'] = 6  # Length of major ticks on the x-axis
plt.rcParams['ytick.major.size'] = 6  # Length of major ticks on the y-axis
plt.rcParams['xtick.minor.size'] = 4   # Length of minor ticks on the x-axis
plt.rcParams['ytick.minor.size'] = 4   # Length of minor ticks on the y-axis

axiswidth = 0.8

#Directory of data for ML
directory_data = "C:/.../ML_results"

#outdir
outdir = "C:/ML_results"

#Logfile
logfile = open(f'{outdir}/ML_logfile.txt','w')
now = datetime.datetime.now()
logfile.write(f'Logfile Machine Learning - | Intialised: {now} \n')

#%%
#==============================================================================
                   
#              Data Processing & Feature Engineering  - EXP 1 and 3

#==============================================================================

#Load data
data = pd.read_csv(f'{directory_data}/stalling_data_tower.csv')     
data = data.fillna(0)

#Seperate datetime into a date and time column
df = data
df['Time_correct'] = [str(x).split(' ')[-1] for x in df['time']]  
df['Date'] = [str(x).split(' ')[-0] for x in df['time']]

#--------------------------------------------------------------------------------

#Prepare datasets for One-Class Classification
df_OneClass_1 = df[df["target"] == 1]

X_OneClass_1 = df_OneClass_1.drop(['time','target','Time_correct','Date'],axis=1)
y_OneClass_1 = df_OneClass_1['target'].values
OneClass_1_feature_names = X_OneClass_1.columns

df_OneClass = df[df["target"] == 0]
X_OneClass = df_OneClass.drop(['time','target','Time_correct','Date'],axis=1)
y_OneClass = df_OneClass['target'].values
OneClass_feature_names = X_OneClass.columns

#--------------------------------------------------------------------------------

#Prepare datasets for Binary Classification (unbalanced)
X_Binary = df.drop(['time','target','Time_correct','Date'],axis=1)
y_Binary = df['target'].values
Binary_feature_names = X_Binary.columns
unique_dates = df['Date'].unique()

#--------------------------------------------------------------------------------

#Statistical info
Binary_target_per = (len(np.where(df['target'].values==1)[0])  / len(df['target'].values)) * 100
Binary_nontarget_per = (len(np.where(df['target'].values==0)[0])  / len(df['target'].values)) * 100

logfile.write(f"Binary: #-samples: {len(df['target'].values)}, 0-class: {len(np.where(df['target'].values==0)[0])},'\
    f' 1-class: {len(np.where(df['target'].values==1)[0])}, %-0: {round(Binary_nontarget_per,2)} %,'\
        f' %-0: {round(Binary_target_per,2)} %")
logfile.write('\n----------------------------------------------------------------------------------------------------------------------\n')


#%%
#==============================================================================
                   
#                           Feature Extraction

#==============================================================================

#Mutual information
mutual_info = mutual_info_classif(X_Binary,y_Binary) #--> measure of the similarity between two labels of the same data. https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mutual_info_score.html  

mi_df = pd.DataFrame()
mi_df['Features'] = Binary_feature_names
mi_df['Mutual Info'] = mutual_info

#sort dataframe
mi_df = mi_df.sort_values(by='Mutual Info',ascending=False)
mi_df = mi_df.reset_index()

#%%
#Visualise the top XX features 
top = 10     

fig,ax = plt.subplots(dpi=72,figsize=(8.27/1.5, 11.69/2)) #w,h

ax.barh(range(top),mi_df['Mutual Info'].values[:top],align='center',color='k',alpha=0.4)

#statistics
mean = np.nanmean(mi_df['Mutual Info'].values[:top])
std = np.nanstd(mi_df['Mutual Info'].values[:top])
median = np.nanmedian(mi_df['Mutual Info'].values[:top])
q_25 = np.quantile(mi_df['Mutual Info'].values[:top], 0.25)
q_75 = np.quantile(mi_df['Mutual Info'].values[:top], 0.75)


logfile.write(f'Mutual Information - mean: {round(mean,4)}, std: {round(std,4)}, median: {round(median,4)}, 25th quantile: {round(q_25,4)} , 75th quantile: {round(q_75,4)}') 
logfile.write('\n----------------------------------------------------------------------------------------------------------------------\n')

ax.plot([mean]*top,np.linspace(-1,top+0.1,top),color='tab:red',linewidth=axiswidth*2,label='Mean')                      #mean
ax.fill_betweenx(np.linspace(-1,top+0.1,top),mean-std,mean+std,color='tab:red',alpha=0.05,label='Std')        #1x std
ax.plot([median]*top,np.linspace(-1,top+0.1,top),color='k',linewidth=axiswidth*2,label='Median')                        #median
ax.plot([q_25]*top,np.linspace(-1,top+0.1,top),color='tab:blue',linewidth=axiswidth*2,label='Q1')                              #25th quantile
ax.plot([q_75]*top,np.linspace(-1,top+0.1,top),color='tab:purple',linewidth=axiswidth*2,label='Q3')  

#detailing
labels = mi_df['Features'].values[:top]
labels_c = []
for label in labels:
    data = label.split('_')
    if len(data) == 2:
        
        if data[0] == 'F':
            data[0] = 'U'
        
        elif data[0] == 'D':
            data[0] = 'Wdir'
            
        elif data[0] == 'SD':
            data[0] = r'$\sigma_{Wdir}$'
                
        labels_c.append(f'{data[0]} {data[1]}')
    else:
        
        if 'lapse' in data[0]:
            data[0] = r'$\Gamma_{\theta}$'
            data[1] = data[-1]
        
        labels_c.append(f'{data[0]} {data[1]}')

        
ax.set_yticks(range(top), labels_c)

ax.set_ylim(-1,top+0.1)
ax.set_xlim(0,0.08)
ax.legend(loc='lower left', bbox_to_anchor= (0, 1.1), ncol=2,
            borderaxespad=0, frameon=False,markerscale=2) 

ax.xaxis.set_minor_locator(AutoMinorLocator())

ax.spines['top'].set_visible(False)
ax.spines['right'].set_visible(False)

ax.set_xlabel('Mutual Information Score (-)')

plt.savefig(f'{outdir}/Figure_5.png',bbox_inches='tight')
plt.savefig(f'{outdir}/Figure_5.pdf',bbox_inches='tight')
plt.savefig(f'{outdir}/Figure_5.svg',bbox_inches='tight')
plt.show()
plt.close()


#%%
#---------------------------------------------------------------------

#Select the features above a certain variable

#Find indices above a certain value (e.g., mean, median, etc.)
threshold = q_25
indices = mi_df[mi_df['Mutual Info'] >= threshold].index

#Get names of features above this index
features_selected = mi_df['Features'].values[indices]
logfile.write(f'n-features selected: {len(features_selected)}, features: {features_selected}') 
logfile.write('\n----------------------------------------------------------------------------------------------------------------------\n')

#--------------------------------------------------------------------
#Update features and labels

X_OneClass = X_OneClass[features_selected]
OneClass_feature_names = X_OneClass.columns

X_OneClass_1 = X_OneClass_1[features_selected]
OneClass_feature_names_1 = X_OneClass_1.columns

X_Binary = X_Binary[features_selected]
Binary_feature_names = X_Binary.columns

#%%
#==============================================================================
                   
#                     Data Normalisation or Scaling

#==============================================================================

#Define a scaler
scaler = MinMaxScaler()

#Scale the data
X_OneClass_scaled = scaler.fit_transform(X_OneClass)
X_OneClass_scaled_1 = scaler.fit_transform(X_OneClass_1)
X_Binary_scaled = scaler.fit_transform(X_Binary)

#%%
#==============================================================================
                   
#                          Hyperameter Grid Formulation

#==============================================================================

#Classifiers
OneClass_classifier = OneClassSVM()

Binary_classifiers = {
    'Logistic Regression': LogisticRegression(),
    'K Neighbors': KNeighborsClassifier(),
    'Gaussian Naive Bayes': GaussianNB(),
    'Random Forest': RandomForestClassifier(),
    'SVM': SVC(),
}

#Grids
gammas = [0.001, 0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1, 10, 100]
nus = [0.001, 0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]

#OneClass
OneClass_parameters = {
    'OneClassSVM': {
        'kernel': ['rbf'],
        'gamma': gammas,
        'nu' : nus,
    }
}

#Binary (un)balanced
Binary_parameters = {
    'Logistic Regression': {
        'C': [0.001, 0.01, 0.1, 1, 10, 100],
        'solver': ['lbfgs', 'liblinear', 'newton-cg', 'sag', 'saga'],
        'max_iter': [10,100,1000,10000,10000]  
    },
    'K Neighbors': {
        'n_neighbors': [1, 5, 10, 20, 50],
        'weights': ['uniform', 'distance']
    },
    'Gaussian Naive Bayes': {},
    'Random Forest': {
        'n_estimators': [50, 100, 200, 500],
        'max_depth': [None, 5, 10, 20, 50],
        'min_samples_split': [2, 5, 10, 20, 50]
    },
    'SVM': {
        'C': [0.001, 0.01, 0.1, 1, 10, 100],
        'kernel': ['linear', 'rbf', 'sigmoid'],
        'gamma': [0.001, 0.01, 0.1, 1, 10, 100]
    }
}

#%%
#==============================================================================
                   
#                        Metric Scorer Formulation

#==============================================================================

#Make precision a scorer
OneClass_scoring_metric = make_scorer(precision_score,pos_label=1)
Binary_scoring_metric = make_scorer(precision_score,pos_label=1)

#%%
#==============================================================================
                   
#                                  OneClass

#==============================================================================

training_sizes = np.round(np.arange(0.3, 1.0, 0.1),2)

kernels = OneClass_parameters['OneClassSVM']['kernel']
gammas = OneClass_parameters['OneClassSVM']['gamma']
nus = OneClass_parameters['OneClassSVM']['nu']

OneClass_Results = pd.DataFrame(columns=["training_size", "kernel", "gamma","nu",'error_train', 'size_train', 'error_test', 'size_test', 'error_0', 'size_0'])

for size in training_sizes:
    size = np.round(size,2)
    
    OneClass_start = datetime.datetime.now()
    logfile.write(f'OneClass SVM Grid Search {int(size*100)}%  training sizes | Intialised: {OneClass_start}')
    logfile.write('\n----------------------------------------------------------------------------------------------------------------------\n')
    for i in range(len(kernels)):
        for j in range(len(gammas)):
            for z in range(len(nus)):
                X_OneClass_train, X_OneClass_test, y_OneClass_train, y_OneClass_test = train_test_split(X_OneClass_scaled_1,y_OneClass_1,test_size=size,random_state=14)

                classifier = OneClassSVM(kernel=kernels[i],gamma=np.round(gammas[j],3),nu=np.round(nus[z],3)).fit(X_OneClass_train)
                
                pred_train = classifier.predict(X_OneClass_train)
                pred_test = classifier.predict(X_OneClass_test)
                pred_0 = classifier.predict(X_OneClass)
                
                size_train = X_OneClass_train.size
                size_test = X_OneClass_test.size
                size_0 = X_OneClass.size
            
                error_train = (pred_train[pred_train == -1].size/X_OneClass_train.size)*100
                error_test = (pred_test[pred_test == -1].size/X_OneClass_test.size)*100
                error_0 = (pred_0[pred_0 == 1].size/X_OneClass.size)*100
                
                new_row = {"training_size" : size, "kernel" : kernels[i], "gamma" : gammas[j],"nu" : nus[z],
                           'error_train': error_train, 'size_train' : size_train, 'error_test' : error_test, 'size_test': size_test, 'error_0' : error_0, 'size_0' : size_0}
                OneClass_Results = OneClass_Results.append(new_row, ignore_index=True)
                OneClass_Results.to_csv(f'{outdir}/OneClassResults.txt', sep='\t', index=False)


#%%

#==============================================================================
                   
#                        Binary Classification (balanced)

#==============================================================================
max_points = np.arange(1,7,1)
complete_index = np.array(df.index)

for maximum in max_points:
    maximum = int(maximum)
    extended_indices = []
    for date in unique_dates:
        df_sub = df[df["Date"] == date]
        df_sub = df_sub[df_sub["target"] == 1]
        
        index = np.array(df_sub.index)
        original_start = index[0]
        original_end = index[-1]
        
        for i in range(1, maximum + 1):  # Start from 1, not 0
            start = original_start - i
            end = original_end + i
            
            # Insert values both at the beginning and end of the index array
            index = np.insert(index, 0, start)
            index = np.insert(index, len(index), end)
        
        fixed_index = []
        for value in index:
            if value in complete_index:
                fixed_index.append(value)

        extended_indices.extend(fixed_index)

    Balanced_df = df.loc[extended_indices]
    
    X_Binary_balanced = Balanced_df.drop(['time','target','Time_correct','Date'],axis=1)
    y_Binary_balanced = Balanced_df['target'].values
    Binary_balanced_feature_names = X_Binary_balanced.columns

    #Statistical info
    Binary_balanced_Target = Balanced_df['target'].values
    Binary_balanced_Target_len = len(Binary_balanced_Target)
    Binary_balanced_Target_num = int(np.sum(Binary_balanced_Target))
    Binary_balanced_nonTarget_num = Binary_balanced_Target_len  - Binary_balanced_Target_num 
    Binary_balanced_Target_per = (Binary_balanced_Target_num  / Binary_balanced_Target_len) * 100
    Binary_balanced_nonTarget_per = (Binary_balanced_nonTarget_num  / Binary_balanced_Target_len) * 100

    logfile.write(f'Binary_balanced: n-points before/after: {maximum}, #-samples: {Binary_balanced_Target_len}, 0-class: {Binary_balanced_nonTarget_num},'\
        f' 1-class: {Binary_balanced_Target_num}, %-0: {round(Binary_balanced_nonTarget_per,2)} %,'\
            f' %-1: {round(Binary_balanced_Target_per,2)} %')
    logfile.write('\n----------------------------------------------------------------------------------------------------------------------\n')

    #Select features
    X_Binary_balanced = X_Binary_balanced[features_selected]
    Binary_balanced_feature_names = X_Binary_balanced.columns
    
    #Scale
    X_Binary_balanced_scaled = scaler.fit_transform(X_Binary_balanced)
    
    Binary_balanced_parameters = Binary_parameters
    Binary_balanced_classifiers = Binary_classifiers
    
    #Fitting
    for size in training_sizes:
        best_estimators = {}
        results = {}
        for key in Binary_classifiers:
            param_grid = Binary_balanced_parameters[key]
            
            locals()[f'{key}_grid'] = GridSearchCV(Binary_balanced_classifiers[key],param_grid,refit=True,verbose=0,scoring=Binary_scoring_metric)         #execute grid search
            
            Binary_balanced_start = datetime.datetime.now()
            logfile.write(f'Binary_balanced {key} | Intialised: {Binary_balanced_start}')
            logfile.write('\n----------------------------------------------------------------------------------------------------------------------\n')
            
            X_Binary_balanced_train, X_Binary_balanced_test, y_Binary_balanced_train, y_Binary_balanced_test = train_test_split(X_Binary_balanced_scaled,y_Binary_balanced,test_size=size)                                                                  
            locals()[f'{key}_grid'].fit(X_Binary_balanced_train,y_Binary_balanced_train)                                                                              #execute fitting of the data
            
            Binary_balanced_end = datetime.datetime.now()
            logfile.write(f'Binary_balanced {key} | Finalised: {Binary_balanced_end} | Duration: {Binary_balanced_end - Binary_balanced_start}')
            logfile.write('\n----------------------------------------------------------------------------------------------------------------------\n')
            
            best_estimators[key] = locals()[f'{key}_grid'].best_estimator_
            results[key] = locals()[f'{key}_grid'].cv_results_ 
            
            Binary_balanced_predictions = locals()[f'{key}_grid'].predict(X_Binary_balanced_test)

            #Metrics
            accuracy = accuracy_score(y_Binary_balanced_test,Binary_balanced_predictions)
            precision, recall, f1, support = precision_recall_fscore_support(y_Binary_balanced_test,Binary_balanced_predictions)

            # Create a DataFrame to store the metrics
            Binary_balanced_metrics= pd.DataFrame({
                'Class': ['Class 0', 'Class 1'],
                'Accuracy' : accuracy,
                'Precision': precision,
                'Recall': recall,
                'F1-Score': f1,
                'Support': support
            })

            Binary_balanced_metrics.to_csv(f'{outdir}/Binary_balancedMetrics_best_{key}_max{maximum}_{int(size*100)}%.txt', sep='\t', index=False)


        #Export best_estimators and grid_results
        # Write the JSON strings to text files
        with open(f'{outdir}/Binary_balanced_best_estimators_max{maximum}_{int(size*100)}%.txt', "w") as best_estimator_file:
            best_estimator_file.write(json.dumps(best_estimators,default=str, indent=4))

        with open(f'{outdir}/Binary_balanced_results_max{maximum}_{int(size*100)}%.txt', "w") as results_file:
            results_file.write(json.dumps(results,default=str, indent=4))