'''
This .py-file is used to prepare the data for machine learning as used in the publication: 

"Advancing Data Quality Assurance with Machine Learning: A Case Study on Wind Vane Stalling Detection"

Authors: Vincent S. de Feiter, Jessica M.I. Strickland and Irene Garcia Marti | Contact: vincent.defeiter@wur.nl 

'''

#%%
#------------------------------------------------------------------------------
#                   I M P O R T   L I B R A R I E S 
#------------------------------------------------------------------------------  
import pandas as pd
import os
import os.path
import logging
import sys
from os import path
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
import matplotlib.colors as mcolors
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.ticker import AutoMinorLocator
import matplotlib.patheffects as PathEffects
from datetime import datetime

import metpy.calc as mpcalc
from metpy.cbook import get_test_data
from metpy.plots import Hodograph, SkewT, colortables
from metpy.units import units
from datetime import datetime, timedelta
import requests
from pathlib import Path


#------------------------------------------------------------------------------
#                               D E F I N E   P A T H S 
#------------------------------------------------------------------------------
#%%

#Location where ALL data is stored
rootdir_all = ".../../Data/..."

#Output directory
outdir = ".../../Machine_Learning_prep/"

#------------------------------------------------------------------------------
#                              S E T T I N G S 
#------------------------------------------------------------------------------
#%%

#Use LATEX font
plt.rcParams['text.usetex'] = True
plt.rcParams['text.latex.preamble'] = r'\usepackage{bm}'
plt.rcParams.update({
    "font.family": "serif",
    "font.serif": ["Times New Roman"],
})

fontsizes=90
pads = 30
pads_border = 1
axiswidth = 3
linewidth = axiswidth

sharey = True                  #share y-axis
plot_grid = False              #add grid

z_list = [10.,20.,40.,80.,140.,200.]

#------------------------------------------------------------------------------
#                        L O A D  &  P R O C E S S  D A T A
#------------------------------------------------------------------------------
#%%

#Load datasets
tower_meteo = xr.open_dataset(f"{rootdir_all}cesar_tower_meteo_lb1_t10_v1.0.nc")                  #general dataset with wind data
surface_meteo = xr.open_dataset(f"{rootdir_all}cesar_surface_meteo_lb1_t10_v1.0.nc")              #general dataset with meteorological obs (primarily pressure)
surface_flux = xr.open_dataset(f"{rootdir_all}cesar_surface_flux_lb1_t10_v1.0.nc")                #general dataset with fluxes
surface_radiation = xr.open_dataset(f"{rootdir_all}cesar_surface_radiation_lb1_t10_v1.0.nc")      #general dataset with radiation

#Pre-process (round to 10 minute intervals and cut equally)
tower_meteo['time'] = pd.to_datetime(tower_meteo.coords['time'].values).round('10T')
surface_meteo['time'] = pd.to_datetime(surface_meteo.coords['time'].values).round('10T')
surface_flux['time'] = pd.to_datetime(surface_flux.coords['time'].values).round('10T')
surface_radiation['time'] = pd.to_datetime(surface_radiation.coords['time'].values).round('10T')

tower_meteo = tower_meteo.sel(time=slice('2001-01-01', '2022-12-31'))
surface_meteo = surface_meteo.sel(time=slice('2001-01-01', '2022-12-31'))
surface_flux = surface_flux.sel(time=slice('2001-01-01', '2022-12-31'))
surface_radiation = surface_radiation.sel(time=slice('2001-01-01', '2022-12-31'))


#Interpolated pressure, potential temperature and FSI
sp = surface_meteo.P0

pressure_complete = np.zeros((len(sp),len(z_list)))
pot_T_complete = np.zeros((len(sp),len(z_list)))
FSI_complete = np.zeros((len(sp),len(z_list)))

z_list = np.array(z_list)
    
#pressure
lapse_rate = -0.125 #hPa/m
for n in range(len(z_list)):
    pressure_complete[:,n] = sp + (lapse_rate * z_list[n])

#Potential temperature
R = 287.05
p0 = 1000
cp = 1004.67

for n in range(len(z_list)):
    T = tower_meteo.TA.sel(z=z_list[n]).values
    p = pressure_complete[:,n]

    pot_T_complete[:,n] = T * (p0 / p) ** (R / cp)

#FSI
for n in range(len(z_list)):
    T = tower_meteo.TA.sel(z=z_list[n]).values
    TA002 = surface_meteo.TA002.values
    TD002 = surface_meteo.TD002.values
    uz = tower_meteo.F.sel(z=z_list[n]).values

    FSI_complete[:,n] = 2*(TA002 - TD002) + 2*(TA002 - T) + uz

#Data Arrays
pressure = xr.DataArray(pressure_complete,coords=[sp.time,z_list],dims=['time','z'])
pot_T = xr.DataArray(pot_T_complete,coords=[sp.time,z_list],dims=['time','z'])
FSI = xr.DataArray(FSI_complete,coords=[sp.time,z_list],dims=['time','z'])

lapse_rate = pot_T.differentiate('z')

#------------------------------------------------------------------------------
#                      I D E N T I F Y   S T A L L I N G 
#------------------------------------------------------------------------------
#%%

#Identify stalling at 10 m
    
#Make selection
sel_tower_meteo = tower_meteo.sel(z=10)
sel_surface_meteo = surface_meteo              #only for surface pressure

#Get indices
diff = sel_tower_meteo.diff(dim='time')

zeros = np.where(diff.D.values == 0)[0]     #based on wind direction

index_zero = []        
#Only look into it when the there are multiple indices where gradient == 0. 
if len(zeros) > 2:
    
    #get differences amongst the points,  only find points consecutive to a distance of 1 or 3
    for i in range(len(zeros)):
        
        # Check the previous and next zeros
        prev_index = zeros[i - 1] if i > 0 else None
        next_index = zeros[i + 1] if i < len(zeros) - 1 else None
        
        if (prev_index is not None and zeros[i] - prev_index <= 3) or (next_index is not None and next_index - zeros[i] <= 3):
            index_zero.append(zeros[i])


#Remove where windspeed is above 6 m/s
wind_speed = sel_tower_meteo.F.astype(float).values[index_zero]
index_below = np.where(wind_speed < 6)[0]
index_zero = np.array(index_zero)[index_below]

#Stalling info
stalling_number = np.zeros((len(sel_tower_meteo.time),1)).flatten()
stalling_number[np.array(index_zero)] = 1    

#General - select only days where stalling occurs
time = sel_tower_meteo.time[index_zero]
dates = time.dt.strftime('%Y-%m-%d').values.astype(str)
years = time.dt.strftime('%Y').values.astype(int)
drop_years = [2011,2012,2022]
idx = np.where(~np.isin(years,drop_years))[0]
dates = dates[idx]
unique_dates = np.unique(dates)

#make selection of the data
indices = xr.DataArray(stalling_number,coords=[sel_tower_meteo.time.values],dims=['time'])

#Get indices and datetime
sel_stalling_indices = indices.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).values
sel_time = sel_tower_meteo.time.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).values

#Get variables - Surface
ps = sel_surface_meteo.P0.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
RAIN = sel_surface_meteo.RAIN.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
RDUR = sel_surface_meteo.RDUR.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
SWD = sel_surface_meteo.SWD.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
TA002 = sel_surface_meteo.TA002.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values - 273.15
TD002 = sel_surface_meteo.TD002.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values - 273.15
RH002 = sel_surface_meteo.RH002.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
Q002 = sel_surface_meteo.Q002.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
TA000 = sel_surface_meteo.TA000.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values - 273.15
F010 = sel_surface_meteo.F010.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
D010 = sel_surface_meteo.D010.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
ZMA = sel_surface_meteo.ZMA.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
PWA = sel_surface_meteo.PWA.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values

#Get variables - Tower
heights = [10.,20.,40.,80.,140.,200.]

for z in heights:
        
        #Make selection
        sel_tower_meteo = tower_meteo.sel(z=z)
        
        #Get variables
        locals()[f'F_{int(z)}m'] = sel_tower_meteo.F.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
        locals()[f'PF_{int(z)}m'] = sel_tower_meteo.PF.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
        locals()[f'MF_{int(z)}m'] = sel_tower_meteo.MF.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
        locals()[f'SF_{int(z)}m'] = sel_tower_meteo.SF.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
        locals()[f'D_{int(z)}m'] = sel_tower_meteo.D.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
        locals()[f'SD_{int(z)}m'] = sel_tower_meteo.SD.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
        locals()[f'TA_{int(z)}m'] = sel_tower_meteo.TA.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values - 273.15
        locals()[f'TD_{int(z)}m'] = sel_tower_meteo.TD.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values - 273.15

#Get variables - Flux
USTED = surface_flux.USTED.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
FCED = surface_flux.FCED.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
HSON = surface_flux.HSON.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
LEED = surface_flux.LEED.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
FG0 = surface_flux.FG0.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
QNBAL = surface_flux.QNBAL.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
USTAB = surface_flux.USTAB.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
USTPR = surface_flux.USTPR.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
WTV = surface_flux.WTV.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
FFSN = surface_flux.FFSN.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values

#Get variables - Radiation
SWD = surface_radiation.SWD.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
SWU = surface_radiation.SWU.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
LWD = surface_radiation.LWD.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
LWU = surface_radiation.LWU.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
SWDZ = surface_radiation.SWDZ.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
SWUZ = surface_radiation.SWUZ.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
SWDTP = surface_radiation.SWDTP.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
LWDTP = surface_radiation.LWDTP.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values
TSLWU = surface_radiation.TSLWU.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).astype(float).values

#Get variables - FSI
FSI_10m = FSI.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=10.0).astype(float).values
FSI_20m = FSI.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=20.0).astype(float).values
FSI_40m = FSI.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=40.0).astype(float).values
FSI_80m = FSI.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=80.0).astype(float).values
FSI_140m = FSI.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=140.0).astype(float).values
FSI_200m = FSI.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=200.0).astype(float).values

#Get variables - Lapse rate
lapse_rate_10m = lapse_rate.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=10.0).astype(float).values
lapse_rate_20m = lapse_rate.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=20.0).astype(float).values
lapse_rate_40m = lapse_rate.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=40.0).astype(float).values
lapse_rate_80m = lapse_rate.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=80.0).astype(float).values
lapse_rate_140m = lapse_rate.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=140.0).astype(float).values
lapse_rate_200m = lapse_rate.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=200.0).astype(float).values

#Get variables - Potential temperature
pot_T_10m = pot_T.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=10.0).astype(float).values
pot_T_20m = pot_T.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=20.0).astype(float).values
pot_T_40m = pot_T.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=40.0).astype(float).values
pot_T_80m = pot_T.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=80.0).astype(float).values
pot_T_140m = pot_T.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=140.0).astype(float).values
pot_T_200m = pot_T.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=200.0).astype(float).values

#Get variables - Pressure
pressure_10m = pressure.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=10.0).astype(float).values
pressure_20m = pressure.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=20.0).astype(float).values
pressure_40m = pressure.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=40.0).astype(float).values
pressure_80m = pressure.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=80.0).astype(float).values
pressure_140m = pressure.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=140.0).astype(float).values
pressure_200m = pressure.sel(time=indices['time'].dt.strftime('%Y-%m-%d').isin(unique_dates)).sel(z=200.0).astype(float).values


#------------------------------------------------------------------------------

#%%

#Export (.csv) - NOTE: We only select the necessary variables, but can be extended whenever needed. 

df = pd.DataFrame()

#General
df['time'] = sel_time
df['target'] = sel_stalling_indices

#Surface
# df['ps'] = ps
# df['RAIN'] = RAIN
# df['RDUR'] = RDUR
# df['SWD'] = SWD
df['TA002'] = TA002
df['TD002'] = TD002
df['RH002'] = RH002
df['Q002'] = Q002
df['TA000'] = TA000
# df['F010'] = F010
# df['D010'] = D010
df['ZMA'] = ZMA
# df['PWA'] = PWA

#Tower
for z in heights:
    df[f'F_{int(z)}m'] = locals()[f'F_{int(z)}m']
    # df[f'PF_{int(z)}m'] = locals()[f'PF_{int(z)}m']
    # df[f'MF_{int(z)}m'] = locals()[f'MF_{int(z)}m']
    # df[f'SF_{int(z)}m'] = locals()[f'SF_{int(z)}m']
    df[f'D_{int(z)}m'] = locals()[f'D_{int(z)}m']
    df[f'SD_{int(z)}m'] = locals()[f'SD_{int(z)}m']
    df[f'TA_{int(z)}m'] = locals()[f'TA_{int(z)}m']
    df[f'TD_{int(z)}m'] = locals()[f'TD_{int(z)}m']

#Flux
# df['USTED'] = USTED
# df['FCED'] = FCED
# df['HSON'] = HSON
# df['LEED'] = LEED
# df['FG0'] = FG0
# df['QNBAL'] = QNBAL
# df['USTAB'] = USTAB
# df['USTPR'] = USTPR
# df['WTV'] = WTV
# df['FFSN'] = FFSN

#Radiation
# df['SWD'] = SWD
# df['SWU'] = SWU
# df['LWD'] = LWD
# df['LWU'] = LWU
# df['SWDZ'] = SWDZ
# df['SWUZ'] = SWUZ
# df['SWDTP'] = SWDTP
# df['LWDTP'] = LWDTP
# df['TSLWU'] = TSLWU

#FSI
df['FSI_10m'] = FSI_10m
df['FSI_20m'] = FSI_20m
df['FSI_40m'] = FSI_40m
df['FSI_80m'] = FSI_80m
df['FSI_140m'] = FSI_140m
df['FSI_200m'] = FSI_200m

#Lapse rate
df['lapse_rate_10m'] = lapse_rate_10m
df['lapse_rate_20m'] = lapse_rate_20m
df['lapse_rate_40m'] = lapse_rate_40m
df['lapse_rate_80m'] = lapse_rate_80m
df['lapse_rate_140m'] = lapse_rate_140m
df['lapse_rate_200m'] = lapse_rate_200m

#Potential temperature
df['pot_T_10m'] = pot_T_10m
df['pot_T_20m'] = pot_T_20m
df['pot_T_40m'] = pot_T_40m
df['pot_T_80m'] = pot_T_80m
df['pot_T_140m'] = pot_T_140m
df['pot_T_200m'] = pot_T_200m

#Pressure
df['pressure_10m'] = pressure_10m
df['pressure_20m'] = pressure_20m
df['pressure_40m'] = pressure_40m
df['pressure_80m'] = pressure_80m
df['pressure_140m'] = pressure_140m
df['pressure_200m'] = pressure_200m

#Export
df.to_csv(f'{outdir}/stalling_data_tower.csv',index=False)