'''
This .py-file is used to generate the figures as used in the publication: 

"Advancing Data Quality Assurance with Machine Learning: A Case Study on Wind Vane Stalling Detection"

Authors: Vincent S. de Feiter, Jessica M.I. Strickland and Irene Garcia Marti | Contact: vincent.defeiter@wur.nl 

'''
#%%
#------------------------------------------------------------------------------
#                   I M P O R T   L I B R A R I E S 
#------------------------------------------------------------------------------  
import pandas as pd
import os
import os.path
import logging
import sys
from os import path
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
import matplotlib.colors as mcolors
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.ticker import AutoMinorLocator
from matplotlib.ticker import MultipleLocator
import matplotlib.patheffects as PathEffects
from datetime import datetime

import metpy.calc as mpcalc
from metpy.cbook import get_test_data
from metpy.plots import Hodograph, SkewT, colortables
from metpy.units import units
from datetime import datetime, timedelta
import requests
from pathlib import Path
import matplotlib.image as mpimg
from itertools import cycle

from sklearn.feature_selection import mutual_info_classif           #Mutual information --> Feature Extraction. 
from sklearn.model_selection import GridSearchCV                    #Gridsearch (for Model Tuning)

#Selected models
from sklearn.svm import OneClassSVM                                 #OneClass Support Vector Machine: binary approach, only with the majority class. 
from sklearn.svm import SVC                                         #Support Vector Machine: binary approach, with the majority and minority class. 
from sklearn.linear_model import LogisticRegression                 #Logistic Regression
from sklearn.neighbors import KNeighborsClassifier                  #K-Neighbors
from sklearn.naive_bayes import GaussianNB                          #Naïve Bayes
from sklearn.ensemble import RandomForestClassifier                 #Random Forest Classifier (multiple decison trees)

from sklearn.model_selection import train_test_split                #splitting the data in train and test data
from sklearn.preprocessing import MinMaxScaler, StandardScaler      #Feature Scaling
from sklearn.metrics import classification_report                   #Metrics: classification report. Build a text report showing the main classification metrics.
from sklearn.metrics import accuracy_score 
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import make_scorer
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import f1_score
from sklearn.metrics import confusion_matrix
from collections import Counter

from matplotlib.colors import ListedColormap, BoundaryNorm
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.colors as mcolors

#------------------------------------------------------------------------------
#                               D E F I N E   P A T H S 
#------------------------------------------------------------------------------
#%%

#Location where ALL data is stored
rootdir_all = "E:/..../Data ALL/"

#Location where ML results are stored
results_ML = "C:/Users/.../ML_results/"

#Location where KNMI weather chart is stored
image_root = "C:/Users/.../Figures/"

#Location where ML data are stored
directory_ML_data = "C:/Users/.../ML_data/"

#Output directory
outdir = "C:/Users/.../Figures/"

#------------------------------------------------------------------------------
#                              S E T T I N G S 
#------------------------------------------------------------------------------
#%%

#Use LATEX font
plt.rcParams['text.usetex'] = True
plt.rcParams['text.latex.preamble'] = r'\usepackage{bm}'
plt.rcParams.update({
    "font.family": "serif",
    "font.serif": ["Arial"],
})


fontsize = 15
plt.rcParams['font.size'] = fontsize

# # Update the font sizes for ylabel, xlabel, and tick labels
plt.rcParams['axes.labelsize'] = fontsize   # X and Y labels

# # Optionally, you can scale other elements, such as legend size, here
plt.rcParams['legend.fontsize'] = fontsize
plt.rcParams['axes.titlepad'] = 20   # Title padding
plt.rcParams['axes.labelpad'] = 20   # Axis label padding
plt.rcParams['xtick.major.pad']= 5 
plt.rcParams['ytick.major.pad']= 5

plt.rcParams['axes.edgecolor'] = 'k'  # Color of the plot border
plt.rcParams['axes.linewidth'] = 1  # Width of the plot border (spines)
plt.rcParams['xtick.major.width'] = 1  # Width of major ticks on the x-axis
plt.rcParams['ytick.major.width'] = 1  # Width of major ticks on the y-axis
plt.rcParams['xtick.minor.width'] = 1  # Width of minor ticks on the x-axis
plt.rcParams['ytick.minor.width'] = 1  # Width of minor ticks on the y-axis
plt.rcParams['xtick.major.size'] = 6  # Length of major ticks on the x-axis
plt.rcParams['ytick.major.size'] = 6  # Length of major ticks on the y-axis
plt.rcParams['xtick.minor.size'] = 4   # Length of minor ticks on the x-axis
plt.rcParams['ytick.minor.size'] = 4   # Length of minor ticks on the y-axis

axiswidth = 0.8


z_list = [10.,20.,40.,80.,140.,200.]

#------------------------------------------------------------------------------
#                        L O A D  &  P R O C E S S  D A T A
#------------------------------------------------------------------------------
#%%

#Load datasets (2001 - 2022)
tower_meteo = xr.open_dataset(f"{rootdir_all}cesar_tower_meteo_lb1_t10_v1.0.nc")                  #general dataset with wind data
surface_meteo = xr.open_dataset(f"{rootdir_all}cesar_surface_meteo_lb1_t10_v1.0.nc")              #general dataset with meteorological obs (primarily pressure)
surface_flux = xr.open_dataset(f"{rootdir_all}cesar_surface_flux_lb1_t10_v1.0.nc")                #general dataset with fluxes
surface_radiation = xr.open_dataset(f"{rootdir_all}cesar_surface_radiation_lb1_t10_v1.0.nc")      #general dataset with radiation

#Pre-process (round to 10 mintue intervals and cut equally)
tower_meteo['time'] = pd.to_datetime(tower_meteo.coords['time'].values).round('10T')
surface_meteo['time'] = pd.to_datetime(surface_meteo.coords['time'].values).round('10T')
surface_flux['time'] = pd.to_datetime(surface_flux.coords['time'].values).round('10T')
surface_radiation['time'] = pd.to_datetime(surface_radiation.coords['time'].values).round('10T')

tower_meteo = tower_meteo.sel(time=slice('2001-01-01', '2022-12-31'))
surface_meteo = surface_meteo.sel(time=slice('2001-01-01', '2022-12-31'))
surface_flux = surface_flux.sel(time=slice('2001-01-01', '2022-12-31'))
surface_radiation = surface_radiation.sel(time=slice('2001-01-01', '2022-12-31'))

#------------------------------------------------------------------------------
#                      I D E N T I F Y   S T A L L I N G 
#------------------------------------------------------------------------------
#%%

#Identify stalling

for n in range(len(z_list)):
    
    #Make selection
    sel_tower_meteo = tower_meteo.sel(z=z_list[n])
    sel_surface_meteo = surface_meteo              #only for surface pressure
    
    #Get indices
    diff = sel_tower_meteo.diff(dim='time')

    zeros = np.where(diff.D.values == 0)[0]     #based on wind direction
    
    index_zero = []        
    #Only look into it when the there are multiple indices where gradient == 0. 
    if len(zeros) > 2:
        
        #get differences amongst the points,  only find points consecutive to a distance of 1 or 3
        for i in range(len(zeros)):
            
            # Check the previous and next zeros
            prev_index = zeros[i - 1] if i > 0 else None
            next_index = zeros[i + 1] if i < len(zeros) - 1 else None
            
            if (prev_index is not None and zeros[i] - prev_index <= 3) or (next_index is not None and next_index - zeros[i] <= 3):
                index_zero.append(zeros[i])


    #Remove where windspeed is above 6 m/s (lower windspeeds)
    wind_speed = sel_tower_meteo.F.astype(float).values[index_zero]
    index_below = np.where(wind_speed < 6)[0]
    index_zero = np.array(index_zero)[index_below]
    
    #Stalling info
    stalling_number = np.zeros((len(sel_tower_meteo.time),1)).flatten()
    stalling_number[np.array(index_zero)] = 1    
    
    #Get variables - Tower
    F = sel_tower_meteo.F.astype(float).values[index_zero]
    PF = sel_tower_meteo.PF.astype(float).values[index_zero]
    MF = sel_tower_meteo.MF.astype(float).values[index_zero]
    SF = sel_tower_meteo.SF.astype(float).values[index_zero]
    D = sel_tower_meteo.D.astype(float).values[index_zero]
    SD = sel_tower_meteo.SD.astype(float).values[index_zero]
    
    TA = sel_tower_meteo.TA.astype(float).values[index_zero] - 273.15
    TD = sel_tower_meteo.TD.astype(float).values[index_zero] - 273.15
    
    #Get variables - Surface
    p = sel_surface_meteo.P0.astype(float).values[index_zero]
    rain = sel_surface_meteo.RAIN.astype(float).values[index_zero]
    SWD = sel_surface_meteo.SWD.astype(float).values[index_zero]
    TA002 = sel_surface_meteo.TA002.astype(float).values[index_zero] - 273.15
    TD002 = sel_surface_meteo.TD002.astype(float).values[index_zero] - 273.15
    RH002 = sel_surface_meteo.RH002.astype(float).values[index_zero]
    Q002 = sel_surface_meteo.Q002.astype(float).values[index_zero]
    F010 = sel_surface_meteo.F010.astype(float).values[index_zero]
    D010 = sel_surface_meteo.D010.astype(float).values[index_zero]
    ZMA = sel_surface_meteo.ZMA.astype(float).values[index_zero]
    
    #General
    time = sel_tower_meteo.time[index_zero]
    
    n_stalling = len(np.unique(time.dt.strftime('%Y-%m-%d').values)) 
    dates = time.dt.strftime('%Y-%m-%d').values.astype(str)
    days = time.dt.strftime('%d').values.astype(int)
    months = time.dt.strftime('%m').values.astype(int)
    years = time.dt.strftime('%Y').values.astype(int)
    
    hours = time.dt.strftime('%H').values.astype(int)
    minutes = time.dt.strftime('%M').values.astype(int)   
    
    #export
    locals()[f'tower_meteo_{int(z_list[n])}m'] = pd.DataFrame({'time':time, 'F':F, 'PF':PF, 
                                                               'MF':MF, 'SF':SF, 'D':D, 'SD': SD, 'TA':TA, 'TD':TD, 'p':p, 'rain':rain, 'SWD':SWD, 'TA002':TA002, 'TD002':TD002, 'RH002':RH002, 'Q002':Q002, 'F010':F010, 'D010':D010, 'ZMA':ZMA, 'dates':dates, 'days':days, 'months':months, 'years':years, 'hours':hours, 'minutes':minutes})
    locals()[f'stalling_indices_{int(z_list[n])}m'] = stalling_number

    locals()[f'tower_meteo_{int(z_list[n])}m']['Time_correct'] = [str(x).split('T')[-1] for x in locals()[f'tower_meteo_{int(z_list[n])}m']['time']] 
    locals()[f'tower_meteo_{int(z_list[n])}m']['Date'] = [str(x).split('T')[-0] for x in locals()[f'tower_meteo_{int(z_list[n])}m']['time']]
    
    locals()[f'tower_meteo_{int(z_list[n])}m'].to_csv(f"{outdir}tower_meteo_{int(z_list[n])}m.csv", index=False)

#------------------------------------------------------------------------------
#                      P O S T  P R O C E S S I N G  
#------------------------------------------------------------------------------
#%%

#Yearly percentage: no outliers
year_range = np.arange(2001,2023)
years_to_drop = [2011, 2012, 2022]
year_range_full = year_range
year_range = np.setdiff1d(year_range, years_to_drop)

percentage = []
counts = []
total = []
for n in range(len(z_list)):
    
    data = pd.read_csv(f"{outdir}stalling_{int(z_list[n])}m.csv")

    #Number of days in reference
    ref_date = np.unique(tower_meteo.time.dt.strftime('%Y-%m-%d').values.astype(str))
    ref_years = np.array([datetime.strptime(date, "%Y-%m-%d").year for date in ref_date]).astype(int)

    #Number of days in data
    data_time = np.unique(data['dates'].values.astype(str))
    data_time = np.array([datetime.strptime(date, "%Y-%m-%d").year for date in data_time]).astype(int)

    sub = []
    sub_count = []
    
    for year in year_range:
        
        #check number of occurences in reference
        count = np.count_nonzero(ref_years == year)
        
        #check number of occurences in data
        count_data = np.count_nonzero(data_time == year)
                
        #append percentage
        sub.append((count_data / count) * 100)
        
        #append counts
        sub_count.append(count_data)
    
    total.append(np.sum(np.array(sub_count)))
    percentage.append(sub)
    counts.append(sub_count)

#%%
#Yearly percentage: full range
percentage_full = []
counts_full = []
for n in range(len(z_list)):
    
    data = pd.read_csv(f"{outdir}stalling_{int(z_list[n])}m.csv")

    #Number of days in reference
    ref_date = np.unique(tower_meteo.time.dt.strftime('%Y-%m-%d').values.astype(str))
    ref_years = np.array([datetime.strptime(date, "%Y-%m-%d").year for date in ref_date]).astype(int)

    #Number of days in data
    data_time = np.unique(data['dates'].values.astype(str))
    data_time = np.array([datetime.strptime(date, "%Y-%m-%d").year for date in data_time]).astype(int)

    sub = []
    sub_count = []
    
    for year in year_range_full:
        
        #check number of occurences in reference
        count = np.count_nonzero(ref_years == year)
        
        #check number of occurences in data
        count_data = np.count_nonzero(data_time == year)
                
        #append percentage
        sub.append((count_data / count) * 100)
        
        #append counts
        sub_count.append(count_data)
    
    percentage_full.append(sub)
    counts_full.append(sub_count)

#%%
#------------------------------------------------------------------------------
#                             V I S U A L I S A T I O N
#------------------------------------------------------------------------------

#%%
#=================
#    Figure 2
#=================

#Load data
data = tower_meteo.sel(z=10)
date_list = ['2012-06-13','2012-06-14','2012-06-15']
data_sub  = data.sel(time=data['time'].dt.strftime('%Y-%m-%d').isin(date_list)) 

#Visualisation
fig,ax = plt.subplots(2,1,dpi=72,figsize=(8.27*1.2, 11.69)) #width, height
ax1, ax2 = ax.flatten()

#=================
#    Figure 2a
#=================

#Stalling example
ax3 = ax1.twinx()  

#plot
ax1.plot(np.arange(0,len(data_sub.F.values)),data_sub.F.values,'o',linewidth=axiswidth,markersize=5,color='tab:blue')
ax3.plot(np.arange(0,len(data_sub.D.values)),data_sub.D.values,'o',linewidth=axiswidth,markersize=5,color='tab:green')

cut_data = data['time'].dt.strftime('%Y-%m-%d').values
idx_data = np.where(np.isin(cut_data,date_list))[0]
stalling_ID = locals()[f'stalling_indices_10m'][idx_data]
idx = np.where(stalling_ID == 1)[0]
ax3.plot(np.arange(0,len(data_sub.D.values))[idx],data_sub.D.values[idx],'o',linewidth=axiswidth,markersize=5,color='tab:red')

ax1.fill_between(idx,0,8.2, color='tab:red',alpha=0.1)

#Detailing
ax1.set_ylabel('U (m s$^{-1}$)', color='tab:blue')	
ax1.set_xlabel("Time (UTC)")
ax1.tick_params(axis='x')
ax1.tick_params(axis='y')
ax1.spines['right'].set_visible(False)
ax1.spines['top'].set_visible(False)

ax1.yaxis.set_minor_locator(AutoMinorLocator())
ax1.xaxis.set_minor_locator(AutoMinorLocator())
ax1.tick_params(axis='y', colors='tab:blue')

ax1.set_ylim(0,8.2)

ax3.set_ylabel('Wdir ($^{\circ}$)',color='tab:green')
ax3.tick_params(axis='x')
ax3.tick_params(axis='y')
ax3.spines['top'].set_visible(False)

ax3.yaxis.set_minor_locator(AutoMinorLocator())
ax3.xaxis.set_minor_locator(AutoMinorLocator())

ax3.set_ylim(0,360.2)
ax3.set_yticks(np.arange(0,360+90,90),np.arange(0,360+90,90),color='tab:green')

#Datetime
dates = data_sub.time.dt.strftime('%Y-%m-%d').values
loc1 = dates.tolist().index('2012-06-13')
loc2 = dates.tolist().index('2012-06-14')
loc3 = dates.tolist().index('2012-06-15')

ax1.axvline(loc1, color='k', linestyle='-',linewidth=axiswidth)
ax1.axvline(loc2, color='k', linestyle='-',linewidth=axiswidth)
ax1.axvline(loc3, color='k', linestyle='-',linewidth=axiswidth)

# Round the datetime values to the nearest 10 minutes
rounded_times = data_sub.time.dt.strftime('%H:%M').values
xticks = rounded_times

locs = []
ref = ['00:00','04:00','08:00','12:00','16:00','20:00']
for n in range(len(ref)):
    locs.extend(np.where(xticks == ref[n])[0])

locs = np.sort(np.array(locs)).tolist()
locs.extend([len(dates)])
ref_total = ref*3

ref_total = pd.to_datetime(ref_total).strftime('%H').astype(int).tolist()
ref_total.extend([0])

ax1.set_xticks(locs,ref_total)


ax1.text(loc1+70,8.5, '13-06-2012',ha='center',va='center',color='k')
ax1.text(loc2+70,8.5, '14-06-2012',ha='center',va='center',color='k')
ax1.text(loc3+70,8.5, '15-06-2012',ha='center',va='center',color='k')

#Thresholds
ax1.axhline(0.5, color='k', linestyle='--',linewidth=axiswidth)

ax1.text(-0.05,1.1, 'a)',transform=ax1.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)

ax1.set_xlim(0,locs[-1])
ax3.set_xlim(0,locs[-1])  

#=================
#    Figure 2b
#=================

#KNMI-chart
img = mpimg.imread(f'{image_root}Figure2a.gif')

ax2.imshow(img,interpolation='antialiased')
ax2.axis('off')  

#Detailing
ax2.set_aspect('equal')
ax2.set_xlim(0, img.shape[1])
ax2.set_ylim(img.shape[0], 0)

ax2.text(-0.1,0.95, 'b)',transform=ax2.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)

fig.subplots_adjust(hspace=0.3)


#=================
#    Export
#=================

plt.savefig(f"{outdir}/Figure_2.jpg",bbox_inches='tight')   
plt.savefig(f"{outdir}/Figure_2.svg",bbox_inches='tight',format='svg')
plt.savefig(f"{outdir}/Figure_2.pdf",bbox_inches='tight',format='pdf')

#%%
#=================
#    Figure 3
#=================

#Load data
mean = []
std = []
median = []
q_75 = []
q_25 = []

mean_full = []
std_full = []
median_full = []
q_75_full = []
q_25_full = []

for n in range(len(z_list)):
    
    mean.append(np.mean(percentage[n]))
    std.append(np.std(percentage[n]))
    median.append(np.median(percentage[n]))
    q_75.append(np.percentile(percentage[n],75))
    q_25.append(np.percentile(percentage[n],25))
    
    mean_full.append(np.mean(percentage_full[n]))
    std_full.append(np.std(percentage_full[n]))
    median_full.append(np.median(percentage_full[n]))
    q_75_full.append(np.percentile(percentage_full[n],75))
    q_25_full.append(np.percentile(percentage_full[n],25))
    

data = pd.read_csv(f"{outdir}stalling_10m.csv")
years = data['years'].values

idx = np.where(~np.isin(years, years_to_drop))[0]
months = data['months'].values[idx]
months_full = data['months'].values

idx = np.where(~np.isin(years, years_to_drop))[0]
hours_full = data['hours'].values
hours = data['hours'].values[idx]

#Visualisation
fig,ax = plt.subplots(2,2,dpi=72,figsize=(8.27*1.4, 11.69)) #width, height

ax1, ax2, ax3, ax4 = ax.flatten()

#=================
#    Figure 3a
#=================

# Overview of stalling events with height

ax1.plot(mean,z_list,linewidth=axiswidth,color='k',marker='o',markersize=5)
ax1.fill_betweenx(z_list,q_25,q_75, color='k',alpha=0.08)

#Text - Without outliers
for n in range(len(z_list)):
    if z_list[n] == 20. or z_list[n] == 80. :
        ax1.text(mean[n]-1.3,z_list[n]-2,f"{mean[n]:.2f}",ha='left',va='center',color='k')
    else:
        ax1.text(mean[n]+0.3,z_list[n]-2,f"{mean[n]:.2f}",ha='left',va='center',color='k')

#Detailing
ax1.set_ylabel('z (m)')
ax1.set_xlabel("Vane stalling (\%)")
ax1.tick_params(axis='x')
ax1.tick_params(axis='y')
ax1.spines['right'].set_visible(False)
ax1.spines['top'].set_visible(False)
ax1.set_yticks(np.array(z_list).astype(int),np.array(z_list).astype(int))
ax1.xaxis.set_minor_locator(AutoMinorLocator())

ax1.set_xlim(0,10)
ax1.set_ylim(0,210)

#=================
#    Figure 3b
#=================

ax2.bar(year_range_full,percentage_full[0],linewidth=axiswidth,color='k',width=0.5)

#Detailing
ax2.set_xlabel('Years')
ax2.set_ylabel("Vane stalling (\%)")
ax2.tick_params(axis='x')
ax2.tick_params(axis='y')
ax2.spines['right'].set_visible(False)
ax2.spines['top'].set_visible(False)

ax2.yaxis.set_minor_locator(AutoMinorLocator())
ax2.xaxis.set_minor_locator(MultipleLocator(1))

ax2.set_xticks([2000,2002,2004,2006,2008,2010,2012,2014,2016,2018,2020,2022],['2000','2002','2004','2006','2008','2010','2012','2014','2016','2018','2020','2022'])
ax2.tick_params(axis='x', rotation=90)
ax2.set_xlim(2000,2022.5)

ax2.set_ylim(0,25)

#=================
#    Figure 3c
#=================

# Normalize counts to percentages
counts, bins = np.histogram(months_full, bins=np.arange(1, 14, 1))
percentages = (counts / counts.sum()) * 100

ax3.bar(bins[:-1], percentages,color='k',linewidth=axiswidth,width=0.4,align='center')

#Detailing
ax3.set_xlabel('Months')
ax3.set_ylabel("Vane stalling (\%)")
ax3.tick_params(axis='x')
ax3.tick_params(axis='y')
ax3.spines['right'].set_visible(False)
ax3.spines['top'].set_visible(False)

ax3.yaxis.set_minor_locator(AutoMinorLocator())

ax3.set_xticks(np.arange(1, 13, 1))
ax3.set_xticklabels(['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],rotation=90)


ax3.set_ylim(0,25)

#=================
#    Figure 3d
#=================

counts, bins = np.histogram(hours, bins=np.arange(0, 25, 1))
percentages = (counts / counts.sum()) * 100

#reorder
df = pd.DataFrame({'hours':bins[:-1],'percentages':percentages})

#order between 12 and 12
df1 = df.loc[df['hours'] >= 12]
df2 = df.loc[df['hours'] < 12]

data = df1['percentages'].values
data = np.append(data,df2['percentages'].values)

ax4.bar(np.arange(12,12+24,1), data,color='k',linewidth=axiswidth,width=0.6,align='center')

#Detailing
ax4.set_xlabel('Hours (UTC)')
ax4.set_ylabel("Vane stalling (\%)")
ax4.tick_params(axis='x')
ax4.tick_params(axis='y')
ax4.spines['right'].set_visible(False)
ax4.spines['top'].set_visible(False)

ax4.yaxis.set_minor_locator(AutoMinorLocator())
ax4.xaxis.set_minor_locator(MultipleLocator(1))

ax4.set_xticks(np.arange(12,12+24+6,6),['12','18','0','6','12'])
ax4.set_xlim(12,np.arange(12,12+24+6,6)[-1])
ax4.set_ylim(0,10)

#=================
#    Export
#=================

ax1.text(-0.1,1.1, 'a)',transform=ax1.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)
ax2.text(-0.1,1.1, 'b)',transform=ax2.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)
ax3.text(-0.1,1.1, 'c)',transform=ax3.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)
ax4.text(-0.1,1.1, 'd)',transform=ax4.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)

plt.subplots_adjust(wspace=0.3, hspace=0.5)


plt.savefig(f"{outdir}/Figure_3.jpg",bbox_inches='tight')
plt.savefig(f"{outdir}/Figure_3.svg",bbox_inches='tight',format='svg')
plt.savefig(f"{outdir}/Figure_3.pdf",bbox_inches='tight',format='pdf')

#%%
#------------------------------------------------------------------------------
#             C A L C U L A T E   V A R I A B L E S   ( T O W E R )
#------------------------------------------------------------------------------

#Indices
stalling_indices = locals()[f'stalling_indices_10m']
stalling_indices = np.where(stalling_indices == 1)[0]
stalling_indices_6h = stalling_indices - 6*6

tower_vars = ['D', 'F', 'TA', 'TD', 'TDdepres', 'pot_T', 'FSI']

#Variables
R = 287.05
p0 = 1000
cp = 1004.67

#Interpolated pressure
pressure = surface_meteo.P0
for n in range(len(z_list)):
    lapse_rate = -0.125 #hPa/m
    locals()[f'pressure_{int(z_list[n])}m'] = pressure + lapse_rate * z_list[n]

pot_T_complete = []

for var in tower_vars:
    
    print(var)
    
    locals()[f'{var}_mean'] = []
    locals()[f'{var}_std'] = []
    locals()[f'{var}_median'] = []
    locals()[f'{var}_q_75'] = []
    locals()[f'{var}_q_25'] = []
    
    locals()[f'{var}_mean_full'] = []
    locals()[f'{var}_std_full'] = []
    locals()[f'{var}_median_full'] = []
    locals()[f'{var}_q_75_full'] = []
    locals()[f'{var}_q_25_full'] = []
    
    locals()[f'{var}_mean_6h'] = []
    locals()[f'{var}_std_6h'] = []
    locals()[f'{var}_median_6h'] = []
    locals()[f'{var}_q_75_6h'] = []
    locals()[f'{var}_q_25_6h'] = []
    
    locals()[f'{var}_mean_full_6h'] = []
    locals()[f'{var}_std_full_6h'] = []
    locals()[f'{var}_median_full_6h'] = []
    locals()[f'{var}_q_75_full_6h'] = []
    locals()[f'{var}_q_25_full_6h'] = []


    for n in range(len(z_list)):   
        
        #Make selection 
        sel_tower_meteo = tower_meteo.sel(z=z_list[n])
            
        years_to_drop = [2011, 2012, 2022]
        
        years = sel_tower_meteo.time.dt.strftime('%Y').values.astype(int)[stalling_indices]
        idx = np.where(~np.isin(years, years_to_drop))[0]
        
        if var == 'TDdepres':
            
            #Variables
            TD = sel_tower_meteo.TD.values
            TA = sel_tower_meteo.TA.values
            
            TDdepres = TA - TD
    
        elif var == 'pot_T':
            #Load temperature
            T = sel_tower_meteo.TA.values
            p = locals()[f'pressure_{int(z_list[n])}m']
            
            pot_T = T * (p0 / p) ** (R / cp)
            
            pot_T_complete.append(pot_T)
        
        elif var == 'FSI':
             #FSIz = 2(T2 − Td2) + 2(T2 − Tz) + uz
    
  
            #Variables
            T = sel_tower_meteo.TA.values
            TA002 = surface_meteo.TA002.values
            TD002 = surface_meteo.TD002.values
            uz = sel_tower_meteo.F.values
            
            FSI = 2*(TA002 - TD002) + 2*(TA002 - T) + uz
            
        else:
            locals()[f'{var}'] = sel_tower_meteo[var].values

        #append
        locals()[f'{var}_mean'].append(np.nanmean(locals()[f'{var}'][stalling_indices][idx]))
        locals()[f'{var}_std'].append(np.nanstd(locals()[f'{var}'][stalling_indices][idx]))
        locals()[f'{var}_median'].append(np.nanmedian(locals()[f'{var}'][stalling_indices][idx]))
        locals()[f'{var}_q_75'].append(np.nanpercentile(locals()[f'{var}'][stalling_indices][idx],75))
        locals()[f'{var}_q_25'].append(np.nanpercentile(locals()[f'{var}'][stalling_indices][idx],25))
        
        locals()[f'{var}_mean_full'].append(np.nanmean(locals()[f'{var}'][stalling_indices]))
        locals()[f'{var}_std_full'].append(np.nanstd(locals()[f'{var}'][stalling_indices]))
        locals()[f'{var}_median_full'].append(np.nanmedian(locals()[f'{var}'][stalling_indices]))
        locals()[f'{var}_q_75_full'].append(np.nanpercentile(locals()[f'{var}'][stalling_indices],75))
        locals()[f'{var}_q_25_full'].append(np.nanpercentile(locals()[f'{var}'][stalling_indices],25))
        
        locals()[f'{var}_mean_6h'].append(np.nanmean(locals()[f'{var}'][stalling_indices_6h][idx]))
        locals()[f'{var}_std_6h'].append(np.nanstd(locals()[f'{var}'][stalling_indices_6h][idx]))
        locals()[f'{var}_median_6h'].append(np.nanmedian(locals()[f'{var}'][stalling_indices_6h][idx]))
        locals()[f'{var}_q_75_6h'].append(np.nanpercentile(locals()[f'{var}'][stalling_indices_6h][idx],75))
        locals()[f'{var}_q_25_6h'].append(np.nanpercentile(locals()[f'{var}'][stalling_indices_6h][idx],25))
        
        locals()[f'{var}_mean_full_6h'].append(np.nanmean(locals()[f'{var}'][stalling_indices_6h]))
        locals()[f'{var}_std_full_6h'].append(np.nanstd(locals()[f'{var}'][stalling_indices_6h]))
        locals()[f'{var}_median_full_6h'].append(np.nanmedian(locals()[f'{var}'][stalling_indices_6h]))
        locals()[f'{var}_q_75_full_6h'].append(np.nanpercentile(locals()[f'{var}'][stalling_indices_6h],75))
        locals()[f'{var}_q_25_full_6h'].append(np.nanpercentile(locals()[f'{var}'][stalling_indices_6h],25))

        
# --- calculate lapse rate ---
lapse_rate_mean = []
lapse_rate_std = []
lapse_rate_median = []
lapse_rate_q_75 = []
lapse_rate_q_25 = []

lapse_rate_mean_full = []
lapse_rate_std_full = []
lapse_rate_median_full = []
lapse_rate_q_75_full = []
lapse_rate_q_25_full = []

lapse_rate_mean_6h = []
lapse_rate_std_6h = []
lapse_rate_median_6h = []
lapse_rate_q_75_6h = []
lapse_rate_q_25_6h = []

lapse_rate_mean_full_6h = []
lapse_rate_std_full_6h = []
lapse_rate_median_full_6h = []
lapse_rate_q_75_full_6h = []
lapse_rate_q_25_full_6h = []


pot_T_complete = xr.DataArray(pot_T_complete,coords=[z_list,sel_tower_meteo.time.values],dims=['z','time'])
lapse_rate = pot_T_complete.differentiate('z')

for n in range(len(z_list)):
    
    #Make selection 
    sel_tower_meteo = tower_meteo.sel(z=z_list[n])
         
    years_to_drop = [2011, 2012, 2022]
    
    years = sel_tower_meteo.time.dt.strftime('%Y').values.astype(int)[stalling_indices]
    idx = np.where(~np.isin(years, years_to_drop))[0]
        
    #append data
    lapse_rate_mean.append(np.nanmean(lapse_rate.sel(z=z_list[n]).values[stalling_indices][idx]))
    lapse_rate_std.append(np.nanstd(lapse_rate.sel(z=z_list[n]).values[stalling_indices][idx]))
    lapse_rate_median.append(np.nanmedian(lapse_rate.sel(z=z_list[n]).values[stalling_indices][idx]))
    lapse_rate_q_75.append(np.nanpercentile(lapse_rate.sel(z=z_list[n]).values[stalling_indices][idx],75))
    lapse_rate_q_25.append(np.nanpercentile(lapse_rate.sel(z=z_list[n]).values[stalling_indices][idx],25))
    
    lapse_rate_mean_full.append(np.nanmean(lapse_rate.sel(z=z_list[n]).values[stalling_indices]))
    lapse_rate_std_full.append(np.nanstd(lapse_rate.sel(z=z_list[n]).values[stalling_indices]))
    lapse_rate_median_full.append(np.nanmedian(lapse_rate.sel(z=z_list[n]).values[stalling_indices]))
    lapse_rate_q_75_full.append(np.nanpercentile(lapse_rate.sel(z=z_list[n]).values[stalling_indices],75))
    lapse_rate_q_25_full.append(np.nanpercentile(lapse_rate.sel(z=z_list[n]).values[stalling_indices],25))
    
    lapse_rate_mean_6h.append(np.nanmean(lapse_rate.sel(z=z_list[n]).values[stalling_indices_6h][idx])) 
    lapse_rate_std_6h.append(np.nanstd(lapse_rate.sel(z=z_list[n]).values[stalling_indices_6h][idx]))
    lapse_rate_median_6h.append(np.nanmedian(lapse_rate.sel(z=z_list[n]).values[stalling_indices_6h][idx]))
    lapse_rate_q_75_6h.append(np.nanpercentile(lapse_rate.sel(z=z_list[n]).values[stalling_indices_6h][idx],75))
    lapse_rate_q_25_6h.append(np.nanpercentile(lapse_rate.sel(z=z_list[n]).values[stalling_indices_6h][idx],25))
    
    lapse_rate_mean_full_6h.append(np.nanmean(lapse_rate.sel(z=z_list[n]).values[stalling_indices_6h]))
    lapse_rate_std_full_6h.append(np.nanstd(lapse_rate.sel(z=z_list[n]).values[stalling_indices_6h]))
    lapse_rate_median_full_6h.append(np.nanmedian(lapse_rate.sel(z=z_list[n]).values[stalling_indices_6h]))
    lapse_rate_q_75_full_6h.append(np.nanpercentile(lapse_rate.sel(z=z_list[n]).values[stalling_indices_6h],75))
    lapse_rate_q_25_full_6h.append(np.nanpercentile(lapse_rate.sel(z=z_list[n]).values[stalling_indices_6h],25))


#%%
#------------------------------------------------------------------------------
#             C A L C U L A T E   V A R I A B L E S   (S U R F A C E)
#------------------------------------------------------------------------------

#Indices
stalling_indices = locals()[f'stalling_indices_10m']
stalling_indices = np.where(stalling_indices == 1)[0]
stalling_indices_6h = stalling_indices - 6*6


#Surface variables
surface_vars = ['P0','RAIN','RDUR','SWD','TA002','TD002','RH002','Q002','TA000','F010','D010','ZMA','PWA']

years = surface_meteo.time.dt.strftime('%Y').values.astype(int)[stalling_indices]
idx = np.where(~np.isin(years, years_to_drop))[0]

for var in surface_vars:
    
    locals()[f'{var}_mean'] = np.nanmean(surface_meteo[var].values[stalling_indices][idx])
    locals()[f'{var}_std'] = np.nanstd(surface_meteo[var].values[stalling_indices][idx])
    locals()[f'{var}_median'] = np.nanmedian(surface_meteo[var].values[stalling_indices][idx])
    locals()[f'{var}_q_75'] = np.nanpercentile(surface_meteo[var].values[stalling_indices][idx],75)
    locals()[f'{var}_q_25'] = np.nanpercentile(surface_meteo[var].values[stalling_indices][idx],25)
    
    locals()[f'{var}_mean_full'] = np.nanmean(surface_meteo[var].values[stalling_indices])
    locals()[f'{var}_std_full'] = np.nanstd(surface_meteo[var].values[stalling_indices])
    locals()[f'{var}_median_full'] = np.nanmedian(surface_meteo[var].values[stalling_indices])
    locals()[f'{var}_q_75_full'] = np.nanpercentile(surface_meteo[var].values[stalling_indices],75)
    locals()[f'{var}_q_25_full'] = np.nanpercentile(surface_meteo[var].values[stalling_indices],25)
    
    locals()[f'{var}_mean_6h'] = np.nanmean(surface_meteo[var].values[stalling_indices_6h][idx])
    locals()[f'{var}_std_6h'] = np.nanstd(surface_meteo[var].values[stalling_indices_6h][idx])
    locals()[f'{var}_median_6h'] = np.nanmedian(surface_meteo[var].values[stalling_indices_6h][idx])
    locals()[f'{var}_q_75_6h'] = np.nanpercentile(surface_meteo[var].values[stalling_indices_6h][idx],75)
    locals()[f'{var}_q_25_6h'] = np.nanpercentile(surface_meteo[var].values[stalling_indices_6h][idx],25)
    
    locals()[f'{var}_mean_full_6h'] = np.nanmean(surface_meteo[var].values[stalling_indices_6h])
    locals()[f'{var}_std_full_6h'] = np.nanstd(surface_meteo[var].values[stalling_indices_6h])
    locals()[f'{var}_median_full_6h'] = np.nanmedian(surface_meteo[var].values[stalling_indices_6h])
    locals()[f'{var}_q_75_full_6h'] = np.nanpercentile(surface_meteo[var].values[stalling_indices_6h],75)
    locals()[f'{var}_q_25_full_6h'] = np.nanpercentile(surface_meteo[var].values[stalling_indices_6h],25)
    

#Surface flux
surface_flux_vars = ['USTED','FCED','HSON','LEED','FG0','QNBAL','USTAB','USTPR','WTV','FFSN']

years = surface_flux.time.dt.strftime('%Y').values.astype(int)[stalling_indices]
idx = np.where(~np.isin(years, years_to_drop))[0]

for var in surface_flux_vars:
    
    locals()[f'{var}_mean'] = np.nanmean(surface_flux[var].values[stalling_indices][idx])
    locals()[f'{var}_std'] = np.nanstd(surface_flux[var].values[stalling_indices][idx])
    locals()[f'{var}_median'] = np.nanmedian(surface_flux[var].values[stalling_indices][idx])
    locals()[f'{var}_q_75'] = np.nanpercentile(surface_flux[var].values[stalling_indices][idx],75)
    locals()[f'{var}_q_25'] = np.nanpercentile(surface_flux[var].values[stalling_indices][idx],25)
    
    locals()[f'{var}_mean_full'] = np.nanmean(surface_flux[var].values[stalling_indices])
    locals()[f'{var}_std_full'] = np.nanstd(surface_flux[var].values[stalling_indices])
    locals()[f'{var}_median_full'] = np.nanmedian(surface_flux[var].values[stalling_indices])
    locals()[f'{var}_q_75_full'] = np.nanpercentile(surface_flux[var].values[stalling_indices],75)
    locals()[f'{var}_q_25_full'] = np.nanpercentile(surface_flux[var].values[stalling_indices],25)

    locals()[f'{var}_mean_6h'] = np.nanmean(surface_flux[var].values[stalling_indices_6h][idx])
    locals()[f'{var}_std_6h'] = np.nanstd(surface_flux[var].values[stalling_indices_6h][idx])
    locals()[f'{var}_median_6h'] = np.nanmedian(surface_flux[var].values[stalling_indices_6h][idx])
    locals()[f'{var}_q_75_6h'] = np.nanpercentile(surface_flux[var].values[stalling_indices_6h][idx],75)
    locals()[f'{var}_q_25_6h'] = np.nanpercentile(surface_flux[var].values[stalling_indices_6h][idx],25)
    
    locals()[f'{var}_mean_full_6h'] = np.nanmean(surface_flux[var].values[stalling_indices_6h])
    locals()[f'{var}_std_full_6h'] = np.nanstd(surface_flux[var].values[stalling_indices_6h])
    locals()[f'{var}_median_full_6h'] = np.nanmedian(surface_flux[var].values[stalling_indices_6h])
    locals()[f'{var}_q_75_full_6h'] = np.nanpercentile(surface_flux[var].values[stalling_indices_6h],75)
    locals()[f'{var}_q_25_full_6h'] = np.nanpercentile(surface_flux[var].values[stalling_indices_6h],25)
    
#surface radiation
surface_rad_vars = ['SWD','SWU','LWD','LWU','SWDZ','SWUZ','SWDTP','LWDTP','TSLWU']

years = surface_radiation.time.dt.strftime('%Y').values.astype(int)[stalling_indices]
idx = np.where(~np.isin(years, years_to_drop))[0]

for var in surface_rad_vars:
    
    locals()[f'{var}_mean'] = np.nanmean(surface_radiation[var].values[stalling_indices][idx])
    locals()[f'{var}_std'] = np.nanstd(surface_radiation[var].values[stalling_indices][idx])
    locals()[f'{var}_median'] = np.nanmedian(surface_radiation[var].values[stalling_indices][idx])
    locals()[f'{var}_q_75'] = np.nanpercentile(surface_radiation[var].values[stalling_indices][idx],75)
    locals()[f'{var}_q_25'] = np.nanpercentile(surface_radiation[var].values[stalling_indices][idx],25)
    
    locals()[f'{var}_mean_full'] = np.nanmean(surface_radiation[var].values[stalling_indices])
    locals()[f'{var}_std_full'] = np.nanstd(surface_radiation[var].values[stalling_indices])
    locals()[f'{var}_median_full'] = np.nanmedian(surface_radiation[var].values[stalling_indices])
    locals()[f'{var}_q_75_full'] = np.nanpercentile(surface_radiation[var].values[stalling_indices],75)
    locals()[f'{var}_q_25_full'] = np.nanpercentile(surface_radiation[var].values[stalling_indices],25)
    
    locals()[f'{var}_mean_6h'] = np.nanmean(surface_radiation[var].values[stalling_indices_6h][idx])
    locals()[f'{var}_std_6h'] = np.nanstd(surface_radiation[var].values[stalling_indices_6h][idx])
    locals()[f'{var}_median_6h'] = np.nanmedian(surface_radiation[var].values[stalling_indices_6h][idx])
    locals()[f'{var}_q_75_6h'] = np.nanpercentile(surface_radiation[var].values[stalling_indices_6h][idx],75)
    locals()[f'{var}_q_25_6h'] = np.nanpercentile(surface_radiation[var].values[stalling_indices_6h][idx],25)
    
    locals()[f'{var}_mean_full_6h'] = np.nanmean(surface_radiation[var].values[stalling_indices_6h])
    locals()[f'{var}_std_full_6h'] = np.nanstd(surface_radiation[var].values[stalling_indices_6h])
    locals()[f'{var}_median_full_6h'] = np.nanmedian(surface_radiation[var].values[stalling_indices_6h])
    locals()[f'{var}_q_75_full_6h'] = np.nanpercentile(surface_radiation[var].values[stalling_indices_6h],75)
    locals()[f'{var}_q_25_full_6h'] = np.nanpercentile(surface_radiation[var].values[stalling_indices_6h],25)
    
    

#%%
#------------------------------------------------------------------------------
#                           V I S U A L I S A T I O N 
#------------------------------------------------------------------------------
#%%

#=================
#    Figure 4
#=================

#Visualisation
fig,ax = plt.subplots(2,3,dpi=72,figsize=(8.27*1.4, 11.69),sharey=True) #width, height

ax1, ax2, ax3, ax4, ax5, ax6 = ax.flatten()

#=================
#    Figure 4a
#=================

#Wind speed

ax1.plot(locals()['F_mean'],z_list,linewidth=axiswidth,color='k',marker='o',markersize=5)
ax1.fill_betweenx(z_list,locals()['F_q_25'],locals()['F_q_75'], color='k',alpha=0.08)

ax1.plot(locals()['F_mean_6h'],z_list,linewidth=axiswidth,color='tab:red',marker='o',markersize=5)
ax1.plot(locals()['F_q_25_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')
ax1.plot(locals()['F_q_75_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')  

#Detailing
ax1.set_ylabel('z (m)')
ax1.set_xlabel("U (m s$^{-1}$)")
ax1.tick_params(axis='x')
ax1.tick_params(axis='y')
ax1.spines['right'].set_visible(False)
ax1.spines['top'].set_visible(False)

ax1.set_yticks(np.array(z_list).astype(int),np.array(z_list).astype(int))

ax1.xaxis.set_minor_locator(AutoMinorLocator())
ax1.set_xlim(0,6.5)
ax1.set_ylim(0,210)

#=================
#    Figure 4b
#=================

#Wind direction
ax2.plot(locals()['D_mean'],z_list,linewidth=axiswidth,color='k',marker='o',markersize=5)
ax2.fill_betweenx(z_list,locals()['D_q_25'],locals()['D_q_75'], color='k',alpha=0.08)

ax2.plot(locals()['D_mean_6h'],z_list,linewidth=axiswidth,color='tab:red',marker='o',markersize=5)
ax2.plot(locals()['D_q_25_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')
ax2.plot(locals()['D_q_75_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')  

#Detailing
ax2.set_xlabel("Wdir ($^{\circ}$)")
ax2.tick_params(axis='x')
ax2.tick_params(axis='y')
ax2.spines['right'].set_visible(False)
ax2.spines['top'].set_visible(False)

ax2.set_yticks(np.array(z_list).astype(int),np.array(z_list).astype(int))

ax2.xaxis.set_minor_locator(AutoMinorLocator())

ax2.set_xlim(0,360)
ax2.set_xticks(np.arange(0,360+90,90),np.arange(0,360+90,90))
ax2.set_ylim(0,210)


#=================
#    Figure 4c
#=================

#lapse rate
ax3.plot(locals()['lapse_rate_mean'],z_list,linewidth=axiswidth,color='k',marker='o',markersize=5,label='during stalling')
ax3.fill_betweenx(z_list,locals()['lapse_rate_q_25'],locals()['lapse_rate_q_75'], color='k',alpha=0.08)

ax3.plot(locals()['lapse_rate_mean_6h'],z_list,linewidth=axiswidth,color='tab:red',marker='o',markersize=5,label='6h before stalling')
ax3.plot(locals()['lapse_rate_q_25_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')
ax3.plot(locals()['lapse_rate_q_75_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')  

#Detailing
ax3.set_xlabel(r'$\Gamma_{\theta}$ (K m$^{-1}$)')
ax3.tick_params(axis='x')
ax3.tick_params(axis='y')
ax3.spines['right'].set_visible(False)
ax3.spines['top'].set_visible(False)

ax3.set_yticks(np.array(z_list).astype(int),np.array(z_list).astype(int))

ax3.legend(loc='lower left', bbox_to_anchor= (-1.5, -2), ncol=3,
            borderaxespad=0, frameon=False)

ax3.xaxis.set_minor_locator(AutoMinorLocator())

ax3.set_xlim(0,0.2)
ax3.set_ylim(0,210)
ax3.axvline(0,linestyle='--',color='k',linewidth=axiswidth)

#=================
#    Figure 4d
#=================

#Potential temperature

ax4.plot(locals()['pot_T_mean'],z_list,linewidth=axiswidth,color='k',marker='o',markersize=5)
ax4.fill_betweenx(z_list,locals()['pot_T_q_25'],locals()['pot_T_q_75'], color='k',alpha=0.08)

ax4.plot(locals()['pot_T_mean_6h'],z_list,linewidth=axiswidth,color='tab:red',marker='o',markersize=5)
ax4.plot(locals()['pot_T_q_25_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')
ax4.plot(locals()['pot_T_q_75_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')  

#Detailing
ax4.set_xlabel(r'$\theta$ (K)')
ax4.set_ylabel('z (m)')
ax4.tick_params(axis='x')
ax4.tick_params(axis='y')
ax4.spines['right'].set_visible(False)
ax4.spines['top'].set_visible(False)

ax4.set_yticks(np.array(z_list).astype(int),np.array(z_list).astype(int))

ax4.xaxis.set_minor_locator(AutoMinorLocator())

ax4.set_xlim(275,290)
ax4.set_ylim(0,210)


#=================
#    Figure 4e
#=================

#Dewpoint depression

ax5.plot(locals()['TDdepres_mean'],z_list,linewidth=axiswidth,color='k',marker='o',markersize=5)
ax5.fill_betweenx(z_list,locals()['TDdepres_q_25'],locals()['TDdepres_q_75'], color='k',alpha=0.08)

ax5.plot(locals()['TDdepres_mean_6h'],z_list,linewidth=axiswidth,color='tab:red',marker='o',markersize=5)
ax5.plot(locals()['TDdepres_q_25_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')
ax5.plot(locals()['TDdepres_q_75_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')  

#Detailing
ax5.set_xlabel("Ta - Td ($^{\circ}$C)")
ax5.tick_params(axis='x')
ax5.tick_params(axis='y')
ax5.spines['right'].set_visible(False)
ax5.spines['top'].set_visible(False)

ax5.set_yticks(np.array(z_list).astype(int),np.array(z_list).astype(int))

ax5.xaxis.set_minor_locator(AutoMinorLocator())

ax5.set_xlim(0,10)
ax5.set_ylim(0,210)

#=================
#    Figure 4f
#=================

#FSI

ax6.plot(locals()['FSI_mean'],z_list,linewidth=axiswidth,color='k',marker='o',markersize=5)
ax6.fill_betweenx(z_list,locals()['FSI_q_25'],locals()['FSI_q_75'], color='k',alpha=0.08)

ax6.plot(locals()['FSI_mean_6h'],z_list,linewidth=axiswidth,color='tab:red',marker='o',markersize=5)
ax6.plot(locals()['FSI_q_25_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')
ax6.plot(locals()['FSI_q_75_6h'],z_list,linewidth=axiswidth,color='tab:red',linestyle = '--')  

#Detailing
ax6.set_xlabel("FSI (-)")
ax6.tick_params(axis='x')
ax6.tick_params(axis='y')
ax6.spines['right'].set_visible(False)
ax6.spines['top'].set_visible(False)

ax6.set_yticks(np.array(z_list).astype(int),np.array(z_list).astype(int))

ax6.xaxis.set_minor_locator(AutoMinorLocator())

ax6.set_xlim(-15,25)
ax6.set_xticks(np.arange(-15,25+5,5),np.arange(-15,25+5,5))
ax6.set_ylim(0,210)
ax6.axvline(0,linestyle='--',color='k',linewidth=axiswidth)

#=================
#    Export
#=================

ax1.text(-0.1,1.1, 'a)',transform=ax1.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)
ax2.text(-0.1,1.1, 'b)',transform=ax2.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)
ax3.text(-0.1,1.1, 'c)',transform=ax3.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)
ax4.text(-0.1,1.1, 'd)',transform=ax4.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)
ax5.text(-0.1,1.1, 'e)',transform=ax5.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)
ax6.text(-0.1,1.1, 'f)',transform=ax6.transAxes,fontweight='bold',fontsize=plt.rcParams['font.size']*1.2)

plt.subplots_adjust(hspace=0.5)

plt.savefig(f"{outdir}/Figure_4.jpg",bbox_inches='tight')
plt.savefig(f"{outdir}/Figure_4.svg",bbox_inches='tight',format='svg')
plt.savefig(f"{outdir}/Figure_4.pdf",bbox_inches='tight',format='pdf')

#%%

var_total = []
var_total.extend(surface_vars)
var_total.extend(surface_flux_vars)
var_total.extend(surface_rad_vars)

data = []

# Iterate over each variable in var_total
for var in var_total:
    # Create a dictionary for the current variable
    entry = {
        'variable': var,
        'mean': locals()[f'{var}_mean'],
        'std': locals()[f'{var}_std'],
        'median': locals()[f'{var}_median'],
        'q_25': locals()[f'{var}_q_25'],
        'q_75': locals()[f'{var}_q_75'],
        'mean_6h': locals()[f'{var}_mean_6h'],
        'std_6h': locals()[f'{var}_std_6h'],
        'median_6h': locals()[f'{var}_median_6h'],
        'q_25_6h': locals()[f'{var}_q_25_6h'],
        'q_75_6h': locals()[f'{var}_q_75_6h']
    }
    # Append the dictionary to the list
    data.append(entry)

# Convert the list of dictionaries to a DataFrame
df = pd.DataFrame(data)

df.to_csv(f"{outdir}/surface_variables.csv",index=False)

#%%
#====================================================================================

#Grids
gammas = [0.001]
nus = [0.001]

gammas.extend(np.round(np.arange(0.01, 1.01, np.round(0.01,2)),2))
nus.extend(np.round(np.arange(0.01, 1, np.round(0.01,2)),2))

gammas.extend([10,100])

#OneClass
OneClass_parameters = {
    'OneClassSVM': {
        'kernel': ['rbf'],
        'gamma': gammas,
        'nu' : nus,
    }
}

#Binary (un)balanced
Binary_parameters = {
    'Logistic Regression': {
        'C': [0.001, 0.01, 0.1, 1, 10, 100],
        'solver': ['lbfgs', 'liblinear', 'newton-cg', 'sag', 'saga'],
        'max_iter': [10,100,1000,10000,10000]  
    },
    'K Neighbors': {
        'n_neighbors': [1, 5, 10, 20, 50],
        'weights': ['uniform', 'distance']
    },
    'Gaussian Naive Bayes': {},
    'Random Forest': {
        'n_estimators': [50, 100, 200, 500],
        'max_depth': [None, 5, 10, 20, 50],
        'min_samples_split': [2, 5, 10, 20, 50]
    },
    'SVM': {
        'C': [0.001, 0.01, 0.1, 1, 10, 100],
        'kernel': ['linear', 'rbf', 'sigmoid'],
        'gamma': [0.001, 0.01, 0.1, 1, 10, 100]
    }
}

#%%


#Binary Balanced
training_sizes = np.arange(0.3, 1.0, 0.1)
max_points = np.arange(1,7,1)

classifiers = Binary_parameters.keys()

#read data
for classifier in classifiers:
    
    locals()[f'Accuracy_1_{classifier}'] = []
    locals()[f'Accuracy_0_{classifier}'] = []
    
    locals()[f'Precision_1_{classifier}'] = []
    locals()[f'Precision_0_{classifier}'] = []

    locals()[f'Recall_1_{classifier}'] = []
    locals()[f'Recall_0_{classifier}'] = []  
    
    locals()[f'F1score_1_{classifier}'] = []
    locals()[f'F1score_0_{classifier}'] = []
    
    locals()[f'Support_1_{classifier}'] = []
    locals()[f'Support_0_{classifier}'] = []
    
    for point in max_points:
        
        locals()[f'Accuracy_1_{classifier}_sub'] = []
        locals()[f'Accuracy_0_{classifier}_sub'] = []
        
        locals()[f'Precision_1_{classifier}_sub'] = []
        locals()[f'Precision_0_{classifier}_sub'] = []

        locals()[f'Recall_1_{classifier}_sub'] = []
        locals()[f'Recall_0_{classifier}_sub'] = []  
        
        locals()[f'F1score_1_{classifier}_sub'] = []
        locals()[f'F1score_0_{classifier}_sub'] = []
        
        locals()[f'Support_1_{classifier}_sub'] = []
        locals()[f'Support_0_{classifier}_sub'] = []
        
        for size in training_sizes:
            df = pd.read_csv(f'{results_ML}/Binary_balancedMetrics_best_{classifier}_max{point}_{int(size*100)}%.txt', sep='\t')
            
            locals()[f'Accuracy_1_{classifier}_sub'].append(round(df['Accuracy'].values[1],2))
            locals()[f'Accuracy_0_{classifier}_sub'].append(round(df['Accuracy'].values[0],2))
            
            locals()[f'Precision_1_{classifier}_sub'].append(round(df['Precision'].values[1],2))
            locals()[f'Precision_0_{classifier}_sub'].append(round(df['Precision'].values[0],2))

            locals()[f'Recall_1_{classifier}_sub'].append(round(df['Recall'].values[1],2))
            locals()[f'Recall_0_{classifier}_sub'].append(round(df['Recall'].values[0],2))  
            
            locals()[f'F1score_1_{classifier}_sub'].append(round(df['F1-Score'].values[1],2))
            locals()[f'F1score_0_{classifier}_sub'].append(round(df['F1-Score'].values[0],2))
            
            locals()[f'Support_1_{classifier}_sub'].append(np.array(round(df['Support'].values[1],2))/1000)
            locals()[f'Support_0_{classifier}_sub'].append(np.array(round(df['Support'].values[0],2))/1000)

        locals()[f'Accuracy_1_{classifier}'].append(locals()[f'Accuracy_1_{classifier}_sub'])
        locals()[f'Accuracy_0_{classifier}'].append(locals()[f'Accuracy_0_{classifier}_sub'])
        
        locals()[f'Precision_1_{classifier}'].append(locals()[f'Precision_1_{classifier}_sub'])
        locals()[f'Precision_0_{classifier}'].append(locals()[f'Precision_0_{classifier}_sub'])

        locals()[f'Recall_1_{classifier}'].append(locals()[f'Recall_1_{classifier}_sub'])
        locals()[f'Recall_0_{classifier}'].append(locals()[f'Recall_0_{classifier}_sub']) 
        
        locals()[f'F1score_1_{classifier}'].append(locals()[f'F1score_1_{classifier}_sub'])
        locals()[f'F1score_0_{classifier}'].append(locals()[f'F1score_0_{classifier}_sub'])
        
        locals()[f'Support_1_{classifier}'].append(locals()[f'Support_1_{classifier}_sub'])
        locals()[f'Support_0_{classifier}'].append(locals()[f'Support_0_{classifier}_sub'])

#%%
#=================
#    Figure 6
#=================

#Feature engineering
data = pd.read_csv(f'{directory_ML_data}/stalling_data_tower.csv')     
data = data.fillna(0)

#Seperate datetime into a date and time column
df = data
df['Time_correct'] = [str(x).split(' ')[-1] for x in df['time']]  
df['Date'] = [str(x).split(' ')[-0] for x in df['time']]

#--------------------------------------------------------------------------------

#Prepare datasets for Binary Classification (unbalanced)
X_Binary = df.drop(['time','target','Time_correct','Date'],axis=1)
y_Binary = df['target'].values
Binary_feature_names = X_Binary.columns
unique_dates = df['Date'].unique()


#Mutual information
mutual_info = mutual_info_classif(X_Binary,y_Binary) #--> measure of the similarity between two labels of the same data. https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mutual_info_score.html  

mi_df = pd.DataFrame()
mi_df['Features'] = Binary_feature_names
mi_df['Mutual Info'] = mutual_info

#sort dataframe
mi_df = mi_df.sort_values(by='Mutual Info',ascending=False)
mi_df = mi_df.reset_index()

top = 10     

#statistics
mean = np.nanmean(mi_df['Mutual Info'].values[:top])
std = np.nanstd(mi_df['Mutual Info'].values[:top])
median = np.nanmedian(mi_df['Mutual Info'].values[:top])
q_25 = np.quantile(mi_df['Mutual Info'].values[:top], 0.25)
q_75 = np.quantile(mi_df['Mutual Info'].values[:top], 0.75)

threshold = q_25
indices = mi_df[mi_df['Mutual Info'] >= threshold].index

#Get names of features above this index
features_selected = mi_df['Features'].values[indices]


#Classifiers
OneClass_classifier = OneClassSVM()

Binary_classifiers = {
    'Logistic Regression': LogisticRegression(),
    'K Neighbors': KNeighborsClassifier(),
    'Gaussian Naive Bayes': GaussianNB(),
    'Random Forest': RandomForestClassifier(),
    'SVM': SVC(),
}

#Grids
gammas = [0.001, 0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1, 10, 100]
nus = [0.001, 0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]

#OneClass
OneClass_parameters = {
    'OneClassSVM': {
        'kernel': ['rbf'],
        'gamma': gammas,
        'nu' : nus,
    }
}

#Binary (un)balanced
Binary_parameters = {
    'Logistic Regression': {
        'C': [0.001, 0.01, 0.1, 1, 10, 100],
        'solver': ['lbfgs', 'liblinear', 'newton-cg', 'sag', 'saga'],
        'max_iter': [10,100,1000,10000,10000]  
    },
    'K Neighbors': {
        'n_neighbors': [1, 5, 10, 20, 50],
        'weights': ['uniform', 'distance']
    },
    'Gaussian Naive Bayes': {},
    'Random Forest': {
        'n_estimators': [50, 100, 200, 500],
        'max_depth': [None, 5, 10, 20, 50],
        'min_samples_split': [2, 5, 10, 20, 50]
    },
    'SVM': {
        'C': [0.001, 0.01, 0.1, 1, 10, 100],
        'kernel': ['linear', 'rbf', 'sigmoid'],
        'gamma': [0.001, 0.01, 0.1, 1, 10, 100]
    }
}

#Make precision a scorer
OneClass_scoring_metric = make_scorer(precision_score,pos_label=1)
Binary_scoring_metric = make_scorer(precision_score,pos_label=1)

#Hit and miss

training_size = 0.8
test_size = 0.2

#Data for hit/miss
sub = df.loc[(df['Date'] == '2015-07-09')]
sub_x = sub.drop(['time','target','Time_correct','Date'],axis=1)
sub_y = sub['target'].values
sub_x  = sub_x[features_selected]

scaler = MinMaxScaler()
sub_x = scaler.fit_transform(sub_x)

#Define a scaler
scaler = MinMaxScaler()
X_Binary = X_Binary[features_selected]
Binary_feature_names = X_Binary.columns
X_Binary_scaled = scaler.fit_transform(X_Binary)

#Loop over maxima
maximum_list = np.arange(1,7,1)
complete_index = np.array(df.index)

total_predictions = {}

for key in Binary_classifiers:
    
    overall_predict = []
    
    for maximum in maximum_list: 
        maximum = int(maximum)
        extended_indices = []
        for date in unique_dates:
            df_sub = df[df["Date"] == date]
            df_sub = df_sub[df_sub["target"] == 1]
            
            index = np.array(df_sub.index)
            original_start = index[0]
            original_end = index[-1]
            
            for i in range(1, maximum + 1):  # Start from 1, not 0
                start = original_start - i
                end = original_end + i
                
                # Insert values both at the beginning and end of the index array
                index = np.insert(index, 0, start)
                index = np.insert(index, len(index), end)
            
            fixed_index = []
            for value in index:
                if value in complete_index:
                    fixed_index.append(value)

            extended_indices.extend(fixed_index)

        Balanced_df = df.loc[extended_indices]

        X_Binary_balanced = Balanced_df.drop(['time','target','Time_correct','Date'],axis=1)
        y_Binary_balanced = Balanced_df['target'].values
        Binary_balanced_feature_names = X_Binary_balanced.columns

        #Select features
        X_Binary_balanced = X_Binary_balanced[features_selected]
        Binary_balanced_feature_names = X_Binary_balanced.columns

        #Scale
        X_Binary_balanced_scaled = scaler.fit_transform(X_Binary_balanced)

        Binary_balanced_parameters = Binary_parameters
        Binary_balanced_classifiers = Binary_classifiers

        #Fitting
        best_estimators = {}
        results = {}

        param_grid = Binary_balanced_parameters[key]

        grid = GridSearchCV(Binary_balanced_classifiers[key],param_grid,refit=True,verbose=0,scoring=Binary_scoring_metric)         #execute grid search

        Binary_balanced_start = datetime.datetime.now()

        X_Binary_balanced_train, X_Binary_balanced_test, y_Binary_balanced_train, y_Binary_balanced_test = train_test_split(X_Binary_balanced_scaled,y_Binary_balanced,test_size=test_size)                                                                  
        grid.fit(X_Binary_balanced_train,y_Binary_balanced_train)                                                                              #execute fitting of the data

        Binary_balanced_end = datetime.datetime.now()

        best_estimators[key] = grid.best_estimator_
        results[key]  = grid.cv_results_ 

        Binary_balanced_predictions = grid.predict(X_Binary_balanced_test)

        #Best fit
        best_fit = best_estimators[key]
        
        predict = best_fit.predict(sub_x)
        
        overall_predict.append(np.array(predict))
    
    total_predictions[key] = overall_predict

#Visualisation


fig,ax = plt.subplots(2,1,dpi=72,figsize=(8.27, 11.69/1.5),sharex=True) #width, height

ax1, ax2 = ax.flatten()
ax3 = ax1.twinx()  

ax1.set_position([0.05, 1, 1, 0.45])  # [left, bottom, width, height]
ax3.set_position([0.05, 1, 1, 0.45])  # [left, bottom, width, height]

ax2.set_position([0.05, 0.1, 1, 0.85])  # [left, bottom, width, height]



#=================
#    Figure 6a
#=================

wind_speed = sub['F_10m'].values.astype(float)
wind_dir = sub['D_10m'].values.astype(float)

#Stalling example


#plot
sizer = 5
ax1.plot(np.arange(0,len(wind_speed)),wind_speed,'o',linewidth=axiswidth,markersize=sizer,color='tab:blue')
ax3.plot(np.arange(0,len(wind_dir)),wind_dir,'o',linewidth=axiswidth,markersize=sizer,color='tab:green')

#Add stalling of outlier detection
idx = np.where(sub_y == 1)[0]
ax3.plot(idx,wind_dir[idx],'o',linewidth=axiswidth,markersize=sizer,color='tab:red')
ax3.fill_between(np.arange(idx[0],idx[-1]+2),0,360,facecolor='tab:red',alpha=0.1)

#Detailing
ax1.set_ylabel('U (m s$^{-1}$)', color='tab:blue')	

ax1.tick_params(axis='x')
ax1.tick_params(axis='y')
ax1.spines['right'].set_visible(False)
ax1.spines['top'].set_visible(False)

ax1.yaxis.set_minor_locator(AutoMinorLocator())
ax1.xaxis.set_minor_locator(AutoMinorLocator())
ax1.tick_params(axis='y', colors='tab:blue')

ax1.text(0.5, 1.1, '09-07-2015', horizontalalignment='center',verticalalignment='center', transform=ax1.transAxes,fontsize=plt.rcParams['font.size']*1.2)

ax1.set_ylim(0,8.2)

ax3.set_ylabel('Wdir ($^{\circ}$)',color='tab:green')
ax3.tick_params(axis='x')
ax3.tick_params(axis='y')
ax3.spines['top'].set_visible(False)

ax3.yaxis.set_minor_locator(AutoMinorLocator())
ax3.xaxis.set_minor_locator(AutoMinorLocator())

ax3.set_ylim(0,360.2)
ax3.set_yticks(np.arange(0,360+90,90),np.arange(0,360+90,90),color='tab:green')

#Results
maximum_list = np.arange(1,7,1)

results = np.zeros((len(wind_speed),len(Binary_classifiers)*len(maximum_list)))

count = 0
for key in Binary_classifiers:
    
    predict = total_predictions[key]
    
    for item in range(len(maximum_list)):
        
        predictor = predict[item].astype(float)
        predictor[predictor == 0] = np.nan
        
        results[:,len(maximum_list)*list(Binary_classifiers.keys()).index(key) + item] = predictor + count
    
    count = count + 1

cmap = plt.get_cmap('Set1')
cmap.set_bad(color='white')


levels = np.arange(1, len(Binary_classifiers)+2, 1)
colors = ['tab:green', 'tab:blue', 'tab:red', 'orange', 'tab:purple']
colors = ['tab:red', 'tab:red', 'tab:red', 'tab:red', 'tab:red']

# Create the colormap
cmap = mcolors.LinearSegmentedColormap.from_list('my_own', colors)

# Sample the colormap at certain points and plot it
norm = BoundaryNorm(levels, cmap.N)

cb = ax2.pcolormesh(results.T,cmap=cmap,norm=norm,alpha=0.3)

#alpha and linestyle
edgecolors = np.asarray(cb.get_edgecolor()).copy()
edgecolors[:, 3] = 0.2  
cb.set_edgecolor(edgecolors)
cb.set_linestyle(':')

tick_positions = np.arange(0, len(Binary_classifiers) * len(maximum_list), 1)

ax2.set_yticks(tick_positions+0.5,[1,2,3,4,5,6]*5,minor=True)
ax2.set_yticks(tick_positions,[])
ax2.tick_params(axis='y', which='major', length=0)

ax2.grid(which='major', axis='y', linestyle='--', color='k', linewidth=0.5,alpha=0.5)

#Detailing
ax2.set_xlabel("Time (UTC)")
ax2.set_ylabel('n-points before and after (-)')
ax2.set_ylim(0,30)

ax2.axhline(6, color='k', linestyle='-',linewidth=3)
ax2.axhline(6*2, color='k', linestyle='-',linewidth=3)
ax2.axhline(6*3, color='k', linestyle='-',linewidth=3)
ax2.axhline(6*4, color='k', linestyle='-',linewidth=3)

ax2.text(1.02, 0.1, 'LR', horizontalalignment='left',verticalalignment='center', transform=ax2.transAxes,fontsize=plt.rcParams['font.size']*1.2)
ax2.text(1.02, 0.3, 'KNN', horizontalalignment='left',verticalalignment='center', transform=ax2.transAxes,fontsize=plt.rcParams['font.size']*1.2)
ax2.text(1.02, 0.5, 'GNB', horizontalalignment='left',verticalalignment='center', transform=ax2.transAxes,fontsize=plt.rcParams['font.size']*1.2)
ax2.text(1.02, 0.7, 'RF', horizontalalignment='left',verticalalignment='center', transform=ax2.transAxes,fontsize=plt.rcParams['font.size']*1.2)
ax2.text(1.02, 0.9, 'SVM', horizontalalignment='left',verticalalignment='center', transform=ax2.transAxes,fontsize=plt.rcParams['font.size']*1.2)

time_ref = sub['Time_correct'].values
time_ref = pd.to_datetime(time_ref).strftime('%H')

#tick every 2 hours (interval is 10 minutes)
time_ticks = time_ref[::6*6].astype(int).tolist()
time_ticks.append(time_ticks[0])
ax2.set_xticks(np.linspace(0,144,len(time_ticks)),time_ticks)

#a and b
ax1.text(-0.07, 0.98, 'a)', horizontalalignment='center',verticalalignment='center', transform=ax1.transAxes,fontsize=plt.rcParams['font.size']*1.2)
ax2.text(-0.07, 0.98, 'b)', horizontalalignment='center',verticalalignment='center', transform=ax2.transAxes,fontsize=plt.rcParams['font.size']*1.2)

plt.savefig(f"{outdir}/Figure_6.png", dpi=300, bbox_inches='tight')
plt.savefig(f"{outdir}/Figure_6.pdf", dpi=300, bbox_inches='tight')
plt.savefig(f"{outdir}/Figure_6.svg", dpi=300, bbox_inches='tight')
plt.show()
plt.close()


#%%
#=================
#    Figure 7
#=================
#Visualise
fig,ax = plt.subplots(2,5,dpi=72,figsize=(8.27*1.2, 11.69/2),sharex=True, sharey=True) #w, h
ax1, ax2, ax3, ax4, ax5, ax6, ax7, ax8, ax9, ax10 = ax.flatten()

#Non-stalling
scores = ['Accuracy','Precision','Recall','F1score','Support']
colors = ['tab:blue','tab:red','k','tab:brown','k']
names = ['LR','KNN','GNB','RF','SVM']
for n in range(1,6):
    classifier = list(classifiers)[n-1]    
    count = 0
    for score in scores:
        #Average
        locals()[f'{score}_mean_0'] = np.nanmean(np.array(locals()[f'{score}_0_{classifier}']),axis=1)
        locals()[f'{score}_std_0'] = np.nanstd(np.array(locals()[f'{score}_0_{classifier}']),axis=1)

        #Plot
        
        if score == 'Support':
            
            locals()[f'axtwin_{n}'] = locals()[f'ax{n}'].twinx()
            
            locals()[f'axtwin_{n}'].bar(max_points,np.array(locals()[f'{score}_mean_0'])+np.array(locals()[f'{score}_std_0']),color='k',alpha=0.2,width=0.5,label=f'{score}')
            locals()[f'axtwin_{n}'].bar(max_points,locals()[f'{score}_std_0'],color='k',alpha=1,width=0.5)
        
        elif score == 'F1score':
            locals()[f'ax{n}'].plot(max_points,locals()[f'{score}_mean_0'],linewidth=axiswidth,color=colors[count],marker='o',markersize=5,label='F1-Score')
            locals()[f'ax{n}'].fill_between(max_points,locals()[f'{score}_mean_0']-locals()[f'{score}_std_0'],locals()[f'{score}_mean_0']+locals()[f'{score}_std_0'], color=colors[count],alpha=0.08)
        
    
        else:
            locals()[f'ax{n}'].plot(max_points,locals()[f'{score}_mean_0'],linewidth=axiswidth,color=colors[count],marker='o',markersize=5,label=f'{score}')
            locals()[f'ax{n}'].fill_between(max_points,locals()[f'{score}_mean_0']-locals()[f'{score}_std_0'],locals()[f'{score}_mean_0']+locals()[f'{score}_std_0'], color=colors[count],alpha=0.08)

        
        count = count + 1
        
    locals()[f'ax{n}'].set_title(f'{names[n-1]}',loc='center',fontsize=plt.rcParams['font.size'], y=1.1)
     
    #Specific detailing
    
    #Normal axes
    locals()[f'ax{n}'].tick_params(axis='x')
    locals()[f'ax{n}'].tick_params(axis='y')
    locals()[f'ax{n}'].spines['top'].set_visible(False)
    
    locals()[f'ax{n}'].set_ylim(0,1)
    locals()[f'ax{n}'].set_xlim(0.5,6.5)
    
    
    #Twin axes
    locals()[f'axtwin_{n}'].tick_params(axis='x')
    locals()[f'axtwin_{n}'].tick_params(axis='y')
    
    locals()[f'axtwin_{n}'].spines['top'].set_visible(False)
    if n!=5:
        locals()[f'axtwin_{n}'].set_yticklabels([])
    else:
        pass
    
    locals()[f'axtwin_{n}'].set_ylim(0,10)
    locals()[f'axtwin_{n}'].set_xlim(0.5,6.5)
    
    
#Stalling
scores = ['Accuracy','Precision','Recall','F1score','Support']
colors = ['tab:blue','tab:red','k','tab:brown','k']
sel = 0
for n in range(6,11):
    classifier = list(classifiers)[sel]
    
    count = 0
    for score in scores:
            
            #Average
            locals()[f'{score}_mean_1'] = np.nanmean(np.array(locals()[f'{score}_1_{classifier}']),axis=1)
            locals()[f'{score}_std_1'] = np.nanstd(np.array(locals()[f'{score}_1_{classifier}']),axis=1)
    
            #Plot
            
            if score == 'Support':
                
                locals()[f'axtwin_{n}'] = locals()[f'ax{n}'].twinx()
    
                locals()[f'axtwin_{n}'].bar(max_points,np.array(locals()[f'{score}_mean_1'])+np.array(locals()[f'{score}_std_1']),color='k',alpha=0.2,width=0.5,label=f'{score}')
                locals()[f'axtwin_{n}'].bar(max_points,locals()[f'{score}_std_1'],color='k',alpha=1,width=0.5)
            
            else:
                locals()[f'ax{n}'].plot(max_points,locals()[f'{score}_mean_1'],linewidth=axiswidth,color=colors[count],marker='o',markersize=5,label=f'{score}')
                locals()[f'ax{n}'].fill_between(max_points,locals()[f'{score}_mean_1']-locals()[f'{score}_std_1'],locals()[f'{score}_mean_1']+locals()[f'{score}_std_1'], color=colors[count],alpha=0.08)

            count = count + 1
    
    #Specific detailing
    
    #Normal axes
    locals()[f'ax{n}'].tick_params(axis='x')
    locals()[f'ax{n}'].tick_params(axis='y')
    locals()[f'ax{n}'].spines['top'].set_visible(False)
    
    locals()[f'ax{n}'].set_ylim(0,1)
    locals()[f'ax{n}'].set_xlim(0.5,6.5)
    locals()[f'ax{n}'].set_xticks(max_points,max_points)

    #Twin axes
    locals()[f'axtwin_{n}'].tick_params(axis='x')
    locals()[f'axtwin_{n}'].tick_params(axis='y')
    
    locals()[f'axtwin_{n}'].spines['top'].set_visible(False)
    if n!=10:
        locals()[f'axtwin_{n}'].set_yticklabels([])
    else:
        pass
    
    locals()[f'axtwin_{n}'].set_ylim(0,10)
    locals()[f'axtwin_{n}'].set_xlim(0.5,6.5)
    
    
    sel = sel + 1
    
#Detailing
ax1.legend(loc='lower left', bbox_to_anchor= (0.6, -0.32), ncol=4,
            borderaxespad=0, frameon=False)
    
fig.text(0.5, -0.07, 'n-points before and after (-)', ha='center')
fig.text(0.01, 0.5, 'Metric (-)', va='center', rotation='vertical')

fig.text(-0.005, 0.75, 'a)', va='center', rotation='horizontal',fontsize=plt.rcParams['font.size']*1.2)
fig.text(1.005, 0.75, 'Non-Stalling', va='center', rotation='horizontal',fontsize=plt.rcParams['font.size']*1.2)
fig.text(-0.005, 0.25, 'b)', va='center', rotation='horizontal',fontsize=plt.rcParams['font.size']*1.2)
fig.text(1.005, 0.25, 'Stalling', va='center', rotation='horizontal',fontsize=plt.rcParams['font.size']*1.2)


fig.text(0.99, 0.5, 'Cases (10$^3$)', va='center', rotation=-90)

plt.subplots_adjust(hspace=0.4)


#Export
plt.savefig(f"{outdir}/Figure_7.jpg",bbox_inches='tight')
plt.savefig(f"{outdir}/Figure_7.svg",bbox_inches='tight',format='svg')
plt.savefig(f"{outdir}/Figure_7.pdf",bbox_inches='tight',format='pdf')
plt.show()
plt.close()

#%%
#=================
#    Figure 8
#=================
OneClass = pd.read_csv(f'{results_ML}/OneClassResults.txt', sep='\t')

OneClass_sorted = OneClass.sort_values(by='error_test',ascending=True)

gammas = [0.001, 0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1, 10, 100]
nus = [0.001, 0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]

fig,ax = plt.subplots(dpi=72,figsize=(8.27, 11.69/2)) #w, h


# Define marker cycle
markers = ['o', 's', 'D', '^', 'v', '<', '>']  # Custom marker cycle
marker_cycle = cycle(markers)

# Plotting each gamma
for gamma in gammas:
    line_mean = []
    line_std = []

    for nu in nus:
        df_sub = OneClass.loc[(OneClass['gamma'] == gamma) & (OneClass['nu'] == nu)]
        
        error = df_sub['error_test'].values
        line_mean.append(np.nanmean(error))
        line_std.append(np.nanstd(error))

    ax.plot(nus, line_mean, linewidth=2, 
            marker=next(marker_cycle),  # Use the next marker from the cycle
            markersize=6, 
            label=f'$\gamma$= {gamma}')
    
    # Fill between mean ± std
    ax.fill_between(nus, 
                    np.array(line_mean) - np.array(line_std), 
                    np.array(line_mean) + np.array(line_std), 
                    alpha=0.1)



ax.tick_params(axis='x')
ax.tick_params(axis='y')
ax.spines['top'].set_visible(False)
ax.spines['right'].set_visible(False)

ax.set_xlabel(r'$\nu$ (-)') 
ax.set_ylabel('Classification Error ($\%$)')

ax.set_ylim(0,14)
ax.set_xlim(0,1)

ax.legend(loc='lower left', bbox_to_anchor= (0.02, -0.52), ncol=4,
            borderaxespad=0, frameon=False)

#Export
plt.savefig(f"{outdir}/Figure_8.jpg",bbox_inches='tight')
plt.savefig(f"{outdir}/Figure_8.svg",bbox_inches='tight',format='svg')
plt.savefig(f"{outdir}/Figure_8.pdf",bbox_inches='tight',format='pdf')
plt.show()
plt.close()